import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import CoordinatorResponseCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.peripherals.coordinator.responses.addr_info import AddrInfoResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': CoordinatorMessages.ADDR_INFO,
    'msgid': 'addrInfoTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 64,
    'result': {
        'devNr': 10,
        'did': 42,
    },
    'dpa': b'\x00\x00\x00\x80\x00\x00\x00\x40\x0a\x2a'
}

data_ok_1: dict = {
    'mtype': CoordinatorMessages.ADDR_INFO,
    'msgid': 'addrInfoTest',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'result': {
        'devNr': 5,
        'did': 42,
    },
    'dpa': b'\x00\x00\x00\x80\x02\x04\x00\x23\x05\x2a'
}

data_error: dict = {
    'mtype': CoordinatorMessages.ADDR_INFO,
    'msgid': 'addrInfoTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x00\x80\x04\x04\x01\x23'
}


class AddrInfoResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, AddrInfoResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, AddrInfoResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, AddrInfoResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, AddrInfoResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, AddrInfoResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, AddrInfoResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.COORDINATOR)
        with self.subTest():
            self.assertEqual(response.pcmd, CoordinatorResponseCommands.ADDR_INFO)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, CoordinatorMessages.ADDR_INFO)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            AddrInfoResponse.from_dpa(b'\x00\x00\x00\x80\x00\x00\x00\x40\x0a')

    @parameterized.expand([
        ['from_dpa', data_ok['result']['devNr'], AddrInfoResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['devNr'], AddrInfoResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['devNr'], AddrInfoResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['devNr'], AddrInfoResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_dev_nr(self, _, dev_nr, response: AddrInfoResponse):
        self.assertEqual(response.dev_nr, dev_nr)

    @parameterized.expand([
        ['from_dpa_error', AddrInfoResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', AddrInfoResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_dev_nr_error(self, _, response: AddrInfoResponse):
        self.assertIsNone(response.dev_nr)

    @parameterized.expand([
        ['from_dpa', data_ok['result']['did'], AddrInfoResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['did'], AddrInfoResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['did'], AddrInfoResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['did'], AddrInfoResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_did(self, _, did, response: AddrInfoResponse):
        self.assertEqual(response.did, did)

    @parameterized.expand([
        ['from_dpa_error', AddrInfoResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', AddrInfoResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_did_error(self, _, response: AddrInfoResponse):
        self.assertIsNone(response.did)



import random
import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.coordinator.requests.authorize_bond import AuthorizeBondParams, AuthorizeBondRequest


class AuthorizeBondRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x00\x0d\xff\xff\x01\x00\x00\x00\x00'
        self.json = {
            'mType': 'iqrfEmbedCoordinator_AuthorizeBond',
            'data': {
                'msgId': 'authorizeBondTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'nodes': [
                            {
                                'reqAddr': 1,
                                'mid': 0
                            }
                        ]
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [
            'single_pair',
            [AuthorizeBondParams(req_addr=1, mid=2164554855)],
            b'\x00\x00\x00\x0d\xff\xff\x01\x67\x7c\x04\x81'
        ],
        [
            'three_pairs',
            [
                AuthorizeBondParams(req_addr=1, mid=0),
                AuthorizeBondParams(req_addr=2, mid=2164554855),
                AuthorizeBondParams(req_addr=3, mid=2164554771),
            ],
            b'\x00\x00\x00\x0d\xff\xff\x01\x00\x00\x00\x00\x02\x67\x7c\x04\x81\x03\x13\x7c\x04\x81'
        ]
    ])
    def test_to_dpa(self, _, params, expected):
        request = AuthorizeBondRequest(nodes=params)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [
            'single_pair',
            [AuthorizeBondParams(req_addr=1, mid=2164554855)],
        ],
        [
            'three_pairs',
            [
                AuthorizeBondParams(req_addr=1, mid=0),
                AuthorizeBondParams(req_addr=2, mid=2164554855),
                AuthorizeBondParams(req_addr=3, mid=2164554771),
            ],
        ]
    ])
    def test_to_json(self, _, params: List[AuthorizeBondParams]):
        request = AuthorizeBondRequest(nodes=params, msgid='authorizeBondTest')
        self.json['data']['req']['param']['nodes'] = [{'reqAddr': node.req_addr, 'mid': node.mid} for node in params]
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [
            'single_pair',
            [AuthorizeBondParams(req_addr=1, mid=2164554855)],
            b'\x00\x00\x00\x0d\xff\xff\x01\x67\x7c\x04\x81'
        ],
        [
            'three_pairs',
            [
                AuthorizeBondParams(req_addr=1, mid=0),
                AuthorizeBondParams(req_addr=2, mid=2164554855),
                AuthorizeBondParams(req_addr=3, mid=2164554771),
            ],
            b'\x00\x00\x00\x0d\xff\xff\x01\x00\x00\x00\x00\x02\x67\x7c\x04\x81\x03\x13\x7c\x04\x81'
        ]
    ])
    def test_set_nodes(self, _, params: List[AuthorizeBondParams], dpa):
        nodes = [AuthorizeBondParams(req_addr=1, mid=0)]
        request = AuthorizeBondRequest(nodes=nodes, msgid='authorizeBondTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.nodes = params
        self.json['data']['req']['param']['nodes'] = [{'reqAddr': node.req_addr, 'mid': node.mid} for node in params]
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [1, -1],
        [256, 10],
        [2, 4294967297],
    ])
    def test_invalid_param_members_constructor(self, req_addr, mid):
        with self.assertRaises(RequestParameterInvalidValueError):
            AuthorizeBondParams(req_addr=req_addr, mid=mid)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_req_addr(self, req_addr: int):
        params = AuthorizeBondParams(req_addr=0, mid=1)
        with self.assertRaises(RequestParameterInvalidValueError):
            params.req_addr = req_addr

    @parameterized.expand([
        [-1],
        [4294967297]
    ])
    def test_invalid_mid(self, mid: int):
        params = AuthorizeBondParams(req_addr=0, mid=1)
        with self.assertRaises(RequestParameterInvalidValueError):
            params.mid = mid

    @parameterized.expand([
        [[]],
        [[AuthorizeBondParams(req_addr=random.randint(0, 255), mid=random.randint(0, 0xFFFFFFFF))] * 12]
    ])
    def test_invalid_pair_count(self, params: List[AuthorizeBondParams]):
        with self.assertRaises(RequestParameterInvalidValueError):
            AuthorizeBondRequest(nodes=params)



"""RAM peripheral request messages."""

from .read import ReadRequest
from .write import WriteRequest
from .read_any import ReadAnyRequest

__all__ = [
    'ReadRequest',
    'WriteRequest',
    'ReadAnyRequest'
]



"""UART peripheral response messages."""

from .open import OpenResponse
from .close import CloseResponse
from .write_read import WriteReadResponse
from .clear_write_read import ClearWriteReadResponse

__all__ = [
    'OpenResponse',
    "CloseResponse",
    'WriteReadResponse',
    'ClearWriteReadResponse',
]



"""OS Batch Data parameters."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import Command
from iqrfpy.enums.peripherals import Peripheral
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants


class BatchData:
    """Batch Data class."""

    __slots__ = '_pnum', '_pcmd', '_hwpid', '_pdata'

    def __init__(self, pnum: Union[Peripheral, int], pcmd: Union[Command, int], hwpid: int,
                 pdata: Optional[List[int]] = None):
        """Batch Data constructor.

        Args:
            pnum (Union[Peripheral, int]): Peripheral number.
            pcmd (Union[Command, int]): Peripheral command.
            hwpid (int): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            pdata (List[int], optional): Request data.
        """
        self._validate(pnum=pnum, pcmd=pcmd, hwpid=hwpid, pdata=pdata)
        self._pnum = pnum
        self._pcmd = pcmd
        self._hwpid = hwpid
        self._pdata = pdata if pdata is not None else []

    def _validate(self, pnum: Union[Peripheral, int], pcmd: Union[Command, int], hwpid: int,
                  pdata: Optional[List[int]] = None):
        """Validate batch parameters.

        Args:
            pnum (Union[Peripheral, int]): Peripheral number.
            pcmd (Union[Command, int]): Peripheral command.
            hwpid (int): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            pdata (List[int], optional): Request data.
        """
        self._validate_pnum(pnum)
        self._validate_pcmd(pcmd)
        self._validate_hwpid(hwpid)
        self._validate_pdata(pdata)

    @staticmethod
    def _validate_pnum(pnum: Union[Peripheral, int]):
        """Validate peripheral number parameter.

        Args:
            pnum (Union[Peripheral, int]): Peripheral number.
        Raises:
            RequestParameterInvalidValueError: If pnum is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= pnum <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('PNUM value should be between 0 and 255.')

    @property
    def pnum(self) -> Union[Peripheral, int]:
        """:obj:`Peripheral` or :obj:`int`: Peripheral number.

        Getter and setter.
        """
        return self._pnum

    @pnum.setter
    def pnum(self, value: Union[Peripheral, int]):
        self._validate_pnum(value)
        self._pnum = value

    @staticmethod
    def _validate_pcmd(pcmd: Union[Command, int]):
        """Validate peripheral command parameter.

        Args:
            pcmd (Union[Command, int]): Peripheral command.
        Raises:
            RequestParameterInvalidValueError: If pcmd is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= pcmd <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('PCMD value should be between 0 and 255.')

    @property
    def pcmd(self) -> Union[Command, int]:
        """:obj:`Command` or :obj:`int`: Peripheral command.

        Getter and setter.
        """
        return self._pcmd

    @pcmd.setter
    def pcmd(self, value: Union[Command, int]):
        self._validate_pcmd(value)
        self._pcmd = value

    @staticmethod
    def _validate_hwpid(hwpid: int):
        """Validate hardware profile ID parameter.

        Args:
            hwpid (int): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
        Raises:
            RequestParameterInvalidValueError: If hwpid is less than 0 or greater than 65535.
        """
        if not dpa_constants.HWPID_MIN <= hwpid <= dpa_constants.HWPID_MAX:
            raise RequestParameterInvalidValueError('HWPID value should be between 0 and 65535.')

    @property
    def hwpid(self) -> int:
        """:obj:`int`: Hardware profile ID.

        Getter and setter.
        """
        return self._hwpid

    @hwpid.setter
    def hwpid(self, value: int):
        self._validate_hwpid(value)
        self._hwpid = value

    @staticmethod
    def _validate_pdata(pdata: Optional[List[int]] = None):
        """Validate pdata parameter.

        Args:
            pdata (List[int], optional): Request data.
        """
        if pdata is None:
            return
        if not Common.values_in_byte_range(pdata):
            raise RequestParameterInvalidValueError('PDATA values should be between 0 and 255.')

    @property
    def pdata(self) -> Optional[List[int]]:
        """:obj:`list` of :obj:`int` or :obj:`None`: Request data.

        Getter and setter.
        """
        return self._pdata

    @pdata.setter
    def pdata(self, value: Optional[List[int]] = None):
        self._validate_pdata(value)
        self._pdata = value if value is not None else []

    def to_pdata(self) -> List[int]:
        """Serialize batch data into DPA request pdata.

        Returns:
            :obj:`list` of :obj:`int`: Serialized DPA request pdata.
        """
        data = [self._pnum, self._pcmd, self._hwpid & 0xFF, (self._hwpid >> 8) & 0xFF] + self._pdata
        return [len(data)] + data

    def to_json(self) -> dict:
        """Serialize batch data into JSON API request data.

        Returns:
            :obj:`dict`: Serialized JSON API request data.
        """
        return {
            'pnum': f'{self._pnum:02x}',
            'pcmd': f'{self._pcmd:02x}',
            'hwpid': f'{self._hwpid:04x}',
            'rdata': '.'.join([f'{x:02x}' for x in self._pdata])
        }



import unittest
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.uart.requests.open import OpenRequest
from iqrfpy.utils.dpa import BaudRates


class OpenRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x01\x00\x0C\x00\xff\xff\x03'
        self.json = {
            'mType': 'iqrfEmbedUart_Open',
            'data': {
                'msgId': 'openTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {
                        'baudRate': BaudRates.B9600,
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['baud_rate', BaudRates.B9600, b'\x01\x00\x0c\x00\xff\xff\x03'],
        ['baud_rate', BaudRates.B115200, b'\x01\x00\x0c\x00\xff\xff\x07'],
    ])
    def test_to_dpa(self, _, baud_rate: BaudRates, expected):
        request = OpenRequest(nadr=1, baud_rate=baud_rate)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['baud_rate', BaudRates.B9600],
        ['baud_rate', BaudRates.B115200],
    ])
    def test_to_json(self, _, baud_rate: BaudRates):
        request = OpenRequest(nadr=1, baud_rate=baud_rate, msgid='openTest')
        self.json['data']['req']['param']['baudRate'] = baud_rate
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [BaudRates.B9600, b'\x01\x00\x0c\x00\xff\xff\x03'],
        [BaudRates.B115200, b'\x01\x00\x0c\x00\xff\xff\x07'],
    ])
    def test_set_baud_rate(self, baud_rate: BaudRates, dpa):
        request = OpenRequest(nadr=1, baud_rate=BaudRates.B9600, msgid='openTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.baud_rate = baud_rate
        self.json['data']['req']['param']['baudRate'] = baud_rate
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_params(self, baud_rate: int):
        with self.assertRaises(RequestParameterInvalidValueError):
            OpenRequest(nadr=1, baud_rate=baud_rate)



import unittest
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.node.requests.backup import BackupRequest


class BackupRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x01\x00\x01\x06\xff\xff\x00'
        self.json = {
            'mType': 'iqrfEmbedNode_Backup',
            'data': {
                'msgId': 'backupTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {
                        'index': 0
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['index', 1, 10, b'\x01\x00\x01\x06\xff\xff\x0a'],
        ['index', 1, 170, b'\x01\x00\x01\x06\xff\xff\xaa'],
    ])
    def test_to_dpa(self, _, nadr: int, index: int, expected):
        request = BackupRequest(nadr=nadr, index=index)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['index', 1, 10],
        ['index', 1, 170]
    ])
    def test_to_json(self, _, nadr: int, index: int):
        request = BackupRequest(nadr=nadr, index=index, msgid='backupTest')
        self.json['data']['req']['param']['index'] = index
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [10, b'\x01\x00\x01\x06\xff\xff\x0a'],
        [170, b'\x01\x00\x01\x06\xff\xff\xaa']
    ])
    def test_set_index(self, index, dpa):
        request = BackupRequest(nadr=1, index=0, msgid='backupTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.index = index
        self.json['data']['req']['param']['index'] = index
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_construct_invalid(self, index):
        with self.assertRaises(RequestParameterInvalidValueError):
            BackupRequest(nadr=1, index=index)



from datetime import datetime
import time

from iqrfpy.peripherals.ledr.requests.pulse import PulseRequest
from iqrfpy.iresponse import IResponse
from iqrfpy.transports.mqtt_transport import MqttTransportParams, MqttTransport


def print_formatted(text, data):
    print(f'{text:30s} {data}')


def handler(response: IResponse) -> None:
    print_formatted('Received message at:', datetime.now())
    print_formatted('Message type:', response.mtype)
    print_formatted('Message ID:', response.msgid)
    print_formatted('Response code:', response.rcode)


params = MqttTransportParams(
        host='localhost',
        port=1883,
        client_id='python-lib-test',
        request_topic='Iqrf/DpaRequest',
        response_topic='Iqrf/DpaResponse',
        qos=1,
        keepalive=25
    )
transport = MqttTransport(params=params, callback=handler, auto_init=True)

while True:
    time.sleep(5)
    print('===========')
    print_formatted('Request sent at:', datetime.now())
    request = PulseRequest(nadr=3)
    transport.send(request)
    print_formatted('Request message ID:', request.msgid)



"""Exploration Peripheral Information Data object."""

from dataclasses import dataclass
from typing import Union
from iqrfpy.utils.dpa import PeripheralTypes


@dataclass
class PeripheralInformationData:
    """Exploration Peripheral Information Data class."""

    __slots__ = 'perte', 'pert', 'par1', 'par2'

    def __init__(self, result: dict):
        """Peripheral Information Data constructor.

        Args:
            result (dict): Peripheral enumeration result.
        """
        self.perte: int = result['perTe']
        """Extended peripheral characteristic."""
        self.pert: Union[PeripheralTypes, int] = result['perT']
        """Peripheral type. If the peripheral is not supported or enabled, value is equal
        to :obj:`PERIPHERAL_TYPE_DUMMY`."""
        self.par1 = result['par1']
        """Optional peripheral specific information."""
        self.par2 = result['par2']
        """Optional peripheral specific information."""



"""OS Write TR Configuration request message."""

from typing import Optional, Union
from iqrfpy.enums.commands import OSRequestCommands
from iqrfpy.enums.message_types import OSMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest
from iqrfpy.peripherals.os.os_tr_conf_data import OsTrConfData

__all__ = [
    'WriteTrConfRequest',
    'OsTrConfData'
]


class WriteTrConfRequest(IRequest):
    """OS Write TR Configuration request class."""

    __slots__ = '_configuration', '_rfpgm'

    def __init__(self, nadr: int, configuration: OsTrConfData, rfpgm: int, hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Write TR Configuration request constructor.

        Args:
            nadr (int): Device address.
            configuration (OsTrConfData): TR Configuration.
            rfpgm (int): RFPGM parameters.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate_rfpgm(rfpgm=rfpgm)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.OS,
            pcmd=OSRequestCommands.WRITE_CFG,
            m_type=OSMessages.WRITE_CFG,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._configuration = configuration
        self._rfpgm = rfpgm

    @staticmethod
    def _validate_rfpgm(rfpgm: int) -> None:
        """Validate RFPGM parameter.

        Args:
            rfpgm (int): RFPGM parameters.
        """
        if not dpa_constants.BYTE_MIN <= rfpgm <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('RFPGM should be a value between 0 and 255.')

    @property
    def configuration(self) -> OsTrConfData:
        """:obj:`OsTrConfData`: TR Configuration.

        Getter and setter.
        """
        return self._configuration

    @configuration.setter
    def configuration(self, value: OsTrConfData) -> None:
        self._configuration = value

    @property
    def rfpgm(self) -> int:
        """:obj:`int`: RFPGM parameters.

        Getter and setter.
        """
        return self._rfpgm

    @rfpgm.setter
    def rfpgm(self, value: int) -> None:
        self._validate_rfpgm(value)
        self._rfpgm = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [0] + self._configuration.to_pdata() + [self._rfpgm]
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {
            'checksum': 0,
            'configuration': self._configuration.to_pdata(),
            'rfpgm': self._rfpgm,
        }
        return super().to_json()



"""Sensor Read Sensors request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import SensorRequestCommands
from iqrfpy.enums.peripherals import Standards
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest
from iqrfpy.peripherals.sensor.requests.sensor_written_data import SensorWrittenData

__all__ = [
    'ReadSensorsRequest',
]


class ReadSensorsRequest(IRequest):
    """Sensor Read Sensors request class."""

    __slots__ = '_sensors', '_written_data'

    def __init__(self, nadr: int, sensors: Optional[List[int]] = None,
                 written_data: Optional[List[SensorWrittenData]] = None, hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Read Sensors request constructor.

        Args:
            nadr (int): Device address.
            sensors (List[int], optional): List of sensor indexes to read. If left empty, only the first sensor is read.
            written_data (List[SensorWrittenData], optional): Data to write to sensors.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate_sensors(sensors)
        super().__init__(
            nadr=nadr,
            pnum=Standards.SENSOR,
            pcmd=SensorRequestCommands.READ_SENSORS,
            m_type=None,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._sensors = sensors
        self._written_data = written_data

    @staticmethod
    def _validate_sensors(sensors: Optional[List[int]] = None):
        """Validate sensors parameter.

        Args:
            sensors (List[int], optional): List of sensor indexes to read. If left empty, only the first sensor is read.
        Raises:
            RequestParameterInvalidValueError: If sensors contains more than 32 values or if values are not
                in range from 0 to 31.
        """
        if sensors is not None:
            if len(sensors) == 0:
                return
            if len(sensors) > 32:
                raise RequestParameterInvalidValueError('Sensors length should be at most 32 values.')
            if min(sensors) < 0 or max(sensors) > 31:
                raise RequestParameterInvalidValueError('Sensors values should be between 0 and 31.')

    @property
    def sensors(self) -> Optional[List[int]]:
        """:obj:`list` of :obj:`int` or :obj:`None`: List of sensor indexes to read.
            If left empty, only the first sensor is read.

        Getter and setter.
        """
        return self._sensors

    @sensors.setter
    def sensors(self, value: Optional[List[int]] = None):
        self._validate_sensors(value)
        self._sensors = value

    @property
    def written_data(self) -> Optional[List[SensorWrittenData]]:
        """:obj:`list` of :obj:`SensorWrittenData` or :obj:`None`: Data to write to sensors.

        Getter and setter.
        """
        return self._written_data

    @written_data.setter
    def written_data(self, value: List[SensorWrittenData]):
        self._written_data = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        if self._sensors is not None:
            written_data = []
            if self._written_data is not None:
                for data in self._written_data:
                    written_data += data.to_pdata()
            self._pdata = Common.indexes_to_4byte_bitmap(self._sensors) + written_data
        else:
            self._pdata = None
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        Raises:
            NotImplementedError: Message type not implemented.
        """
        raise NotImplementedError('JSON API request not implemented.')



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import BinaryOutputResponseCommands
from iqrfpy.enums.message_types import BinaryOutputMessages
from iqrfpy.enums.peripherals import Standards
from iqrfpy.peripherals.binaryoutput.responses.enumerate import EnumerateResponse
from iqrfpy.response_factory import ResponseFactory
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': BinaryOutputMessages.ENUMERATE,
    'msgid': 'enumerateTest',
    'nadr': 3,
    'hwpid': 65535,
    'rcode': 0,
    'dpa_value': 71,
    'result': {
        'binOuts': 3
    },
    'dpa': b'\x03\x00\x4b\xbe\xff\xff\x00\x47\x03'
}

data_error: dict = {
    'mtype': BinaryOutputMessages.ENUMERATE,
    'msgid': 'enumerateTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 3,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x4b\xbe\x04\x04\x03\x23'
}


class EnumerateResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        [
            'from_dpa',
            data_ok,
            ResponseFactory.get_response_from_dpa(data_ok['dpa']),
            False
        ],
        [
            'from_json',
            data_ok,
            ResponseFactory.get_response_from_json(generate_json_response(data_ok)),
            True
        ],
        [
            'from_dpa_error',
            data_error,
            ResponseFactory.get_response_from_dpa(data_error['dpa']),
            False
        ],
        [
            'from_json_error',
            data_error,
            ResponseFactory.get_response_from_json(generate_json_response(data_error)),
            True
        ],
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, Standards.BINARY_OUTPUT)
        with self.subTest():
            self.assertEqual(response.pcmd, BinaryOutputResponseCommands.ENUMERATE_OUTPUTS)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, BinaryOutputMessages.ENUMERATE)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    @parameterized.expand([
        [
            'from_dpa',
            data_ok['result']['binOuts'],
            ResponseFactory.get_response_from_dpa(data_ok['dpa']),
        ],
        [
            'from_json',
            data_ok['result']['binOuts'],
            ResponseFactory.get_response_from_json(generate_json_response(data_ok)),
        ],
    ])
    def test_get_count(self, _, count: int, response: EnumerateResponse):
        self.assertEqual(response.count, count)

    @parameterized.expand([
        [
            'from_dpa_error',
            ResponseFactory.get_response_from_dpa(data_error['dpa']),
        ],
        [
            'from_json_error',
            ResponseFactory.get_response_from_json(generate_json_response(data_error)),
        ],
    ])
    def test_get_count_error(self, _, response: EnumerateResponse):
        self.assertIsNone(response.count)



"""Response message class abstraction."""

from abc import ABC, abstractmethod
from typing import List, Optional, Union
from iqrfpy.enums.commands import Command
from iqrfpy.enums.message_types import MessageType
from iqrfpy.enums.peripherals import Peripheral
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.utils.validators import DpaValidator

__all__ = ['IResponse', 'IResponseGetterMixin']


class IResponse(ABC):
    """This class serves as an interface for embedded peripherals and standards response messages.
    This class does not provide any deserialization logic in its factory methods as they simply serve as abstracts.

    Once a factory method is implemented, it should be used to parse responses and create response message objects.
    """

    ASYNC_MSGID = 'async'

    def __init__(self, nadr: int, pnum: Union[Peripheral, int], pcmd: Union[Command, int],
                 hwpid: int = dpa_constants.HWPID_MAX, rcode: int = 0, dpa_value: int = 0,
                 pdata: Optional[List[int]] = None, m_type: Optional[Union[MessageType, str]] = None,
                 msgid: Optional[str] = None, result: Optional[dict] = None):
        self._nadr = nadr
        self._pnum = pnum
        self._pcmd = pcmd
        self._mtype = m_type
        self._hwpid = hwpid
        self._rcode = rcode
        self._dpa_value = dpa_value
        self._pdata = pdata
        self._msgid = msgid
        self._result = result

    @property
    @abstractmethod
    def nadr(self) -> int:
        """:obj:`int`: Device address.

        Getter only.
        """
        return self._nadr

    @property
    @abstractmethod
    def pnum(self) -> Union[Peripheral, int]:
        """:obj:`Peripheral` or :obj:`int`: Peripheral number.

        Getter only.
        """
        return self._pnum

    @property
    @abstractmethod
    def pcmd(self) -> Union[Command, int]:
        """:obj:`Command` or :obj:`int`: Peripheral command.

        Getter only.
        """
        return self._pcmd

    @property
    @abstractmethod
    def mtype(self) -> Optional[Union[MessageType, str]]:
        """:obj:`MessageType` or :obj:`str` or :obj:`None`: Message type.

        Getter only.
        """
        return self._mtype

    @property
    @abstractmethod
    def hwpid(self) -> int:
        """:obj:`int`: Hardware profile ID.

        Getter only.
        """
        return self._hwpid

    @property
    @abstractmethod
    def rcode(self) -> int:
        """:obj:`int`: DPA error code.

        Getter only.
        """
        return self._rcode

    @property
    @abstractmethod
    def dpa_value(self) -> int:
        """:obj:`int`: DPA value.

        Getter only.
        """
        return self._dpa_value

    @property
    @abstractmethod
    def pdata(self) -> Optional[List[int]]:
        """:obj:`list` of :obj:`int`: DPA response data.

        Getter only.
        """
        return self._pdata

    @property
    @abstractmethod
    def result(self) -> Optional[dict]:
        """:obj:`dict` or :obj:`None`: JSON API response data.

        Getter only.
        """
        return self._result

    @property
    @abstractmethod
    def msgid(self) -> Optional[str]:
        """:obj:`str` or :obj:`None`: Message ID.

        Getter only.
        """
        return self._msgid

    @staticmethod
    def validate_dpa_response(data: bytes) -> None:
        """Validate DPA response for base response length.

        Args:
            data (bytes): DPA response.
        """
        DpaValidator.base_response_length(data)

    @classmethod
    @abstractmethod
    def from_dpa(cls, dpa: bytes) -> 'IResponse':
        """Factory method. Parse DPA response into Response object.

        Args:
            dpa (bytes): DPA response.
        """

    @classmethod
    @abstractmethod
    def from_json(cls, json: dict) -> 'IResponse':
        """Factory method. Parse JSON response into Response object.

        Args:
            json (dict): JSON API response.
        """


class IResponseGetterMixin(IResponse):
    """Response mixin."""

    @property
    def nadr(self) -> int:
        """:obj:`int`: Device address.

        Getter only.
        """
        return super().nadr

    @property
    def pnum(self) -> Union[Peripheral, int]:
        """:obj:`Peripheral` or :obj:`int`: Peripheral number.

        Getter only.
        """
        return super().pnum

    @property
    def pcmd(self) -> Union[Command, int]:
        """:obj:`Command` or :obj:`int`: Peripheral command.

        Getter only.
        """
        return super().pcmd

    @property
    def mtype(self) -> Union[MessageType, str]:
        """:obj:`MessageType` or :obj:`str` or :obj:`None`: Message type.

        Getter only.
        """
        return super().mtype

    @property
    def hwpid(self) -> int:
        """:obj:`int`: Hardware profile ID.

        Getter only.
        """
        return super().hwpid

    @property
    def rcode(self) -> int:
        """:obj:`int`: DPA error code.

        Getter only.
        """
        return super().rcode

    def get_rcode_as_string(self) -> str:
        """Return rcode as string.

        Returns:
            str: String representation of rcode.
        """
        return dpa_constants.ResponseCodes.to_string(self.rcode)

    @property
    def dpa_value(self) -> int:
        """:obj:`int`: DPA value.

        Getter only.
        """
        return super().dpa_value

    @property
    def pdata(self) -> Optional[List[int]]:
        """:obj:`list` of :obj:`int`: DPA response data.

        Getter only.
        """
        return super().pdata

    @property
    def result(self) -> Optional[dict]:
        """:obj:`dict` or :obj:`None`: JSON API response data.

        Getter only.
        """
        return super().result

    @property
    def msgid(self) -> Optional[str]:
        """:obj:`str` or :obj:`None`: Message ID.

        Getter only.
        """
        return super().msgid

    @classmethod
    def from_dpa(cls, dpa: bytes) -> IResponse:
        """Factory method. Parse DPA response into Response object.

        Args:
            dpa (bytes): DPA response.
        Raises:
            NotImplementedError: Factory method for mixin not implemented.
        """
        raise NotImplementedError('from_dpa() method not implemented.')

    @classmethod
    def from_json(cls, json: dict) -> IResponse:
        """Factory method. Parse JSON response into Response object.

        Args:
            json (dict): JSON API response.
        Raises:
            NotImplementedError: Factory method for mixin not implemented.
        """
        raise NotImplementedError('from_json() method not implemented.')



"""Node peripheral request messages."""

from .read import ReadRequest
from .remove_bond import RemoveBondRequest
from .backup import BackupRequest
from .restore import RestoreRequest
from .validate_bonds import ValidateBondsRequest, NodeValidateBondsParams

__all__ = [
    'ReadRequest',
    'RemoveBondRequest',
    'BackupRequest',
    'RestoreRequest',
    'ValidateBondsRequest',
    'NodeValidateBondsParams',
]



"""FRC peripheral request messages."""

from iqrfpy.peripherals.frc.frc_params import FrcParams
from .send import SendRequest
from .extra_result import ExtraResultRequest
from .send_selective import SendSelectiveRequest
from .set_frc_params import SetFrcParamsRequest

__all__ = [
    'SendRequest',
    'ExtraResultRequest',
    'SendSelectiveRequest',
    'FrcParams',
    'SetFrcParamsRequest',
]



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import CoordinatorResponseCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.peripherals.coordinator.responses.bond_node import BondNodeResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': CoordinatorMessages.BOND_NODE,
    'msgid': 'bondNodeTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 71,
    'result': {
        'bondAddr': 1,
        'devNr': 2
    },
    'dpa': b'\x00\x00\x00\x84\x00\x00\x00\x47\x01\x02'
}

data_ok_1: dict = {
    'mtype': CoordinatorMessages.BOND_NODE,
    'msgid': 'bondNodeTest',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'result': {
        'bondAddr': 10,
        'devNr': 100
    },
    'dpa': b'\x00\x00\x00\x82\x02\x04\x00\x23\x0a\x64'
}

data_error: dict = {
    'mtype': CoordinatorMessages.BOND_NODE,
    'msgid': 'bondNodeTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 7,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x00\x82\x04\x04\x07\x23'
}


class BondNodeResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, BondNodeResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, BondNodeResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, BondNodeResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, BondNodeResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, BondNodeResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, BondNodeResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.COORDINATOR)
        with self.subTest():
            self.assertEqual(response.pcmd, CoordinatorResponseCommands.BOND_NODE)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, CoordinatorMessages.BOND_NODE)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            BondNodeResponse.from_dpa(b'\x00\x00\x00\x84\x00\x00\x00\x22')

    @parameterized.expand([
        ['from_dpa', data_ok['result']['bondAddr'], BondNodeResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['bondAddr'], BondNodeResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['bondAddr'], BondNodeResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['bondAddr'], BondNodeResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_bond_addr(self, _, bond_addr, response: BondNodeResponse):
        self.assertEqual(response.bond_addr, bond_addr)

    @parameterized.expand([
        ['from_dpa_error', BondNodeResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', BondNodeResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_bond_addr_error(self, _, response: BondNodeResponse):
        self.assertIsNone(response.bond_addr)

    @parameterized.expand([
        ['from_dpa', data_ok['result']['devNr'], BondNodeResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['devNr'], BondNodeResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['devNr'], BondNodeResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['devNr'], BondNodeResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_dev_nr(self, _, dev_nr, response: BondNodeResponse):
        self.assertEqual(response.dev_nr, dev_nr)

    @parameterized.expand([
        ['from_dpa_error', BondNodeResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', BondNodeResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_dev_nr_error(self, _, response: BondNodeResponse):
        self.assertIsNone(response.dev_nr)



"""Coordinator Set Dpa Params request message."""

from enum import IntEnum
from typing import Optional, Union
from iqrfpy.enums.commands import CoordinatorRequestCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['SetDpaParamsRequest', 'DpaParam']


class DpaParam(IntEnum):
    """DpaParam class.

    Parameters for SetDpaParams request.
    """
    LAST_RSSI = 0
    VOLTAGE = 1
    SYSTEM = 2
    USER_SPECIFIED = 3


class SetDpaParamsRequest(IRequest):
    """Coordinator Set Dpa Params request class."""

    __slots__ = ('_dpa_param',)

    def __init__(self, dpa_param: Union[DpaParam, int], hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """SetDpaParams request constructor.

        Args:
            dpa_param (Union[DpaParam, int]): DPA param to set.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate_dpa_param(dpa_param)
        super().__init__(
            nadr=dpa_constants.COORDINATOR_NADR,
            pnum=EmbedPeripherals.COORDINATOR,
            pcmd=CoordinatorRequestCommands.SET_DPA_PARAMS,
            m_type=CoordinatorMessages.SET_DPA_PARAMS,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._dpa_param = dpa_param

    @staticmethod
    def _validate_dpa_param(dpa_param: Union[DpaParam, int]):
        """Validates DPA param parameter.

        Args:
            dpa_param (Union[DpaParam, int]): DPA param to set.
        Raises:
            RequestParameterInvalidValueError: If dpa_param is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= dpa_param <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('DPA param value should be between 0 and 255.')

    @property
    def dpa_param(self) -> Union[DpaParam, int]:
        """:obj:`DpaParam` or :obj:`int`: DPA param.

        Getter and setter.
        """
        return self._dpa_param

    @dpa_param.setter
    def dpa_param(self, value: Union[DpaParam, int]) -> None:
        self._validate_dpa_param(value)
        self._dpa_param = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._dpa_param]
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'dpaParam': self._dpa_param}
        return super().to_json()



"""DPA exploration response messages."""

from .peripheral_enumeration_data import PeripheralEnumerationData
from .peripheral_enumeration import PeripheralEnumerationResponse
from .peripheral_information import PeripheralInformationResponse
from .peripheral_information_data import PeripheralInformationData
from .more_peripherals_information import MorePeripheralsInformationResponse

__all__ = [
    'PeripheralEnumerationData',
    'PeripheralEnumerationResponse',
    'PeripheralInformationData',
    'PeripheralInformationResponse',
    'MorePeripheralsInformationResponse'
]



"""FRC parser module.

Provides methods for parsing and processing FRC data.
"""
import math
from typing import List, Union
from iqrfpy.utils.dpa import FrcCommands


class FrcParser:
    """Class for handling FRC data."""

    @staticmethod
    def values_from_data(data: List[int], extra_data: List[int], frc_command: Union[int, FrcCommands]) -> List[int]:
        """Convert FRC data and extra result data into values collected by FRC commands.

        Note that the method expects raw data and a combined length of 64 bytes from FRC send and extra result requests.

        2 bit FRC commands produces 240 values. 1 byte FRC commands produce 64 values.
        2 byte FRC commands produces 32 values. 4 byte FRC commands produce 16 values.
        The first value should always be a dummy value 0.

        Args:
            data (List[int]): FRC Send request data
            extra_data (List[int]): FRC ExtraResult request data
            frc_command (Union[int, FrcCommands]): FRC command
        Returns:
            :obj:`list` of :obj`int`: Values from collected FRC data
        Raises:
            ValueError: Raised if combined data length is not 64 bytes
        """
        aux = data + extra_data
        if len(aux) != 64:
            raise ValueError('Combined length of FRC data and ExtraResult data should be 64 bytes.')
        if FrcCommands.FRC_2BIT_FROM <= frc_command <= FrcCommands.FRC_2BIT_TO:
            values = []
            for i in range(0, 240):
                mask = 1 << (i % 8)
                idx = math.floor(i / 8)
                val = 0
                if (aux[idx] & mask) != 0:
                    val = 1
                if (aux[idx + 32] & mask) != 0:
                    val |= 2
                values.append(val)
            return values
        if FrcCommands.FRC_1BYTE_FROM <= frc_command <= FrcCommands.FRC_1BYTE_TO:
            return aux
        if FrcCommands.FRC_2BYTE_FROM <= frc_command <= FrcCommands.FRC_2BYTE_TO:
            return [(aux[i + 1] << 8) + aux[i] for i in range(0, len(aux), 2)]
        return [(aux[i + 3] << 24) + (aux[i + 2] << 16) + (aux[i + 1] << 8) + aux[i] for i in range(0, len(aux), 4)]



"""Peripherals module.

This module contains embedded peripherals and standards enums.
These enums are extended with has_value() method for member identification.
"""

from ..utils.enums import IntEnumMember


class Peripheral(IntEnumMember):
    """Base peripherals enum class."""


class EmbedPeripherals(Peripheral):
    """Embedded peripherals enum."""

    COORDINATOR = 0
    NODE = 1
    OS = 2
    EEPROM = 3
    EEEPROM = 4
    RAM = 5
    LEDR = 6
    LEDG = 7
    IO = 9
    THERMOMETER = 10
    UART = 12
    FRC = 13
    EXPLORATION = 255


class Standards(Peripheral):
    """Standards peripherals enum."""

    DALI = 74
    BINARY_OUTPUT = 75
    SENSOR = 94
    LIGHT = 113



"""OS TR Configuration data."""

from typing import List, Optional, Union
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.common import Common
from iqrfpy.utils import dpa as dpa_constants


class OsTrConfData:
    """OS TR Configuration data."""

    __slots__ = '_embedded_peripherals', '_custom_dpa_handler', '_dpa_peer_to_peer', '_routing_off', '_io_setup', \
        '_user_peer_to_peer', '_stay_awake_when_not_bonded', '_std_and_lp_network', '_rf_output_power', \
        '_rf_signal_filter', '_lp_rf_timeout', '_uart_baud_rate', '_alternative_dsm_channel', '_local_frc', \
        '_rf_channel_a', '_rf_channel_b', '_undocumented'

    def __init__(self, embedded_peripherals: Optional[List[Union[EmbedPeripherals, int]]] = None,
                 custom_dpa_handler: bool = False, dpa_peer_to_peer: bool = False, routing_off: bool = False,
                 io_setup: bool = False, user_peer_to_peer: bool = False, stay_awake_when_not_bonded: bool = False,
                 std_and_lp_network: bool = False, rf_output_power: int = 7, rf_signal_filter: int = 5,
                 lp_rf_timeout: int = 6,
                 uart_baud_rate: Union[dpa_constants.BaudRates, int] = dpa_constants.BaudRates.B9600,
                 alternative_dsm_channel: int = 0, local_frc: bool = False, rf_channel_a: int = 52,
                 rf_channel_b: int = 2, undocumented: Optional[List[int]] = None):
        """TR Configuration data constructor.

        Args:
            embedded_peripherals (List[int]): Enabled embedded peripherals.
            custom_dpa_handler (bool): Custom DPA handler in use.
            dpa_peer_to_peer (bool): DPA peer-to-peer enabled.
            routing_off (bool): Node does not route packets in background.
            io_setup (bool): Run IO Setup early during module boot time.
            user_peer_to_peer (bool): Receive peer-to-peer packets and raise PeerToPeer event.
            stay_awake_when_not_bonded (bool): Stay awake during bonding process.
            std_and_lp_network (bool): Control STD+LP network.
            rf_output_power (int): RF output power.
            rf_signal_filter (int): RF signal filter.
            lp_rf_timeout (int): LP RF timeout.
            uart_baud_rate (Union[dpa_constants.BaudRates, int]): UART baud rate.
            alternative_dsm_channel (int): Alternative DSM channel.
            local_frc (bool): Local FRC reception enabled.
            rf_channel_a (int): RF Channel A.
            rf_channel_b (int): RF Channel B.
            undocumented (List[int]): Undocumented data bytes.
        """
        if embedded_peripherals is None:
            embedded_peripherals = []
        if undocumented is None:
            undocumented = [0] * 13
        self._validate(embedded_peripherals=embedded_peripherals, rf_output_power=rf_output_power,
                       rf_signal_filter=rf_signal_filter, lp_rf_timeout=lp_rf_timeout, baud_rate=uart_baud_rate,
                       alternative_dsm=alternative_dsm_channel, rf_channel_a=rf_channel_a, rf_channel_b=rf_channel_b,
                       undocumented=undocumented)
        self._embedded_peripherals = embedded_peripherals
        self._custom_dpa_handler = custom_dpa_handler
        self._dpa_peer_to_peer = dpa_peer_to_peer
        self._routing_off = routing_off
        self._io_setup = io_setup
        self._user_peer_to_peer = user_peer_to_peer
        self._stay_awake_when_not_bonded = stay_awake_when_not_bonded
        self._std_and_lp_network = std_and_lp_network
        self._rf_output_power = rf_output_power
        self._rf_signal_filter = rf_signal_filter
        self._lp_rf_timeout = lp_rf_timeout
        self._uart_baud_rate = uart_baud_rate
        self._alternative_dsm_channel = alternative_dsm_channel
        self._local_frc = local_frc
        self._rf_channel_a = rf_channel_a
        self._rf_channel_b = rf_channel_b
        self._undocumented = undocumented

    def __eq__(self, other: 'OsTrConfData'):
        return self._embedded_peripherals == other._embedded_peripherals and \
            self._custom_dpa_handler == other._custom_dpa_handler and \
            self._dpa_peer_to_peer == other._dpa_peer_to_peer and \
            self._routing_off == other.routing_off and \
            self._io_setup == other._io_setup and \
            self._user_peer_to_peer == other._user_peer_to_peer and \
            self._stay_awake_when_not_bonded == other._stay_awake_when_not_bonded and \
            self._std_and_lp_network == other._std_and_lp_network and \
            self._rf_output_power == other._rf_output_power and \
            self._rf_signal_filter == other._rf_signal_filter and \
            self._lp_rf_timeout == other._lp_rf_timeout and \
            self._uart_baud_rate == other._uart_baud_rate and \
            self._alternative_dsm_channel == other._alternative_dsm_channel and \
            self._local_frc == other._local_frc and \
            self._rf_channel_a == other._rf_channel_a and \
            self._rf_channel_b == other._rf_channel_b and \
            self._undocumented == other._undocumented

    def _validate(self, embedded_peripherals: List[int], rf_output_power: int, rf_signal_filter: int,
                  lp_rf_timeout: int, baud_rate: Union[dpa_constants.BaudRates, int], alternative_dsm: int,
                  rf_channel_a: int, rf_channel_b: int, undocumented: Optional[List[int]] = None):
        self._validate_embedded_peripherals(embedded_peripherals)
        self._validate_rf_output_power(rf_output_power)
        self._validate_rf_signal_filter(rf_signal_filter)
        self._validate_lp_rf_timeout(lp_rf_timeout)
        self._validate_uart_baud_rate(baud_rate)
        self._validate_alternative_dsm_channel(alternative_dsm)
        self._validate_rf_channel_a(rf_channel_a)
        self._validate_rf_channel_b(rf_channel_b)
        self._validate_undocumented(undocumented)

    @staticmethod
    def _validate_embedded_peripherals(embedded_peripherals: List[Union[EmbedPeripherals, int]]) -> None:
        if len(embedded_peripherals) > 32:
            raise RequestParameterInvalidValueError('Embedded peripherals should be at most 32 values.')
        if min(embedded_peripherals, default=0) < 0 or max(embedded_peripherals, default=0) > 31:
            raise RequestParameterInvalidValueError('Embedded peripherals values should be between 0 and 31.')

    @property
    def embedded_peripherals(self) -> List[Union[EmbedPeripherals, int]]:
        """:obj:`list` of :obj:`EmbedPeripherals` or :obj:`int`: Enabled embedded peripherals.

        Getter and setter.
        """
        return self._embedded_peripherals

    @embedded_peripherals.setter
    def embedded_peripherals(self, value: List[Union[EmbedPeripherals, int]]):
        self._validate_embedded_peripherals(embedded_peripherals=value)
        self._embedded_peripherals = value

    @property
    def custom_dpa_handler(self) -> bool:
        """:obj:`bool`: Custom DPA handler in use.

        Getter and setter.
        """
        return self._custom_dpa_handler

    @custom_dpa_handler.setter
    def custom_dpa_handler(self, value: bool):
        self._custom_dpa_handler = value

    @property
    def dpa_peer_to_peer(self) -> bool:
        """:obj:`bool`: DPA peer-to-peer enabled.

        Getter and setter.
        """
        return self._dpa_peer_to_peer

    @dpa_peer_to_peer.setter
    def dpa_peer_to_peer(self, value: bool):
        self._dpa_peer_to_peer = value

    @property
    def routing_off(self) -> bool:
        """:obj:`bool`: Node does not route packets in background.

        Getter and setter.
        """
        return self._routing_off

    @routing_off.setter
    def routing_off(self, value: bool):
        self._routing_off = value

    @property
    def io_setup(self) -> bool:
        """:obj:`bool`: Run IO Setup early during module boot time.

        Getter and setter.
        """
        return self._io_setup

    @io_setup.setter
    def io_setup(self, value: bool):
        self._io_setup = value

    @property
    def user_peer_to_peer(self) -> bool:
        """:obj:`bool`: Receive peer-to-peer packets and raise PeerToPeer event.

        Getter and setter.
        """
        return self._user_peer_to_peer

    @user_peer_to_peer.setter
    def user_peer_to_peer(self, value: bool):
        self._user_peer_to_peer = value

    @property
    def stay_awake_when_not_bonded(self) -> bool:
        """:obj:`bool`: Stay awake during bonding process.

        Getter and setter.
        """
        return self._stay_awake_when_not_bonded

    @stay_awake_when_not_bonded.setter
    def stay_awake_when_not_bonded(self, value: bool):
        self._stay_awake_when_not_bonded = value

    @property
    def std_and_lp_network(self) -> bool:
        """:obj:`bool`: Control STD+LP network.

        Getter and setter.
        """
        return self._std_and_lp_network

    @std_and_lp_network.setter
    def std_and_lp_network(self, value: bool):
        self._std_and_lp_network = value

    @staticmethod
    def _validate_rf_output_power(rf_output_power: int) -> None:
        """Validate rf output power parameter.

        Args:
            rf_output_power (int): RF output power.
        Raises:
            RequestParameterInvalidValueError: If rf_output_power is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= rf_output_power <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('RF output power value should be between 0 and 255.')

    @property
    def rf_output_power(self) -> int:
        """:obj:`int`: RF output power.

        Getter and setter.
        """
        return self._rf_output_power

    @rf_output_power.setter
    def rf_output_power(self, value: int):
        self._validate_rf_output_power(rf_output_power=value)
        self._rf_output_power = value

    @staticmethod
    def _validate_rf_signal_filter(rf_signal_filter: int) -> None:
        """Validate rf signal filter parameter.

        Args:
            rf_signal_filter (int): RF signal filter.
        Raises:
            RequestParameterInvalidValueError: If rf_signal_filter is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= rf_signal_filter <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('RF signal filter value should be between 0 and 255.')

    @property
    def rf_signal_filter(self) -> int:
        """:obj:`int`: RF signal filter.

        Getter and setter.
        """
        return self._rf_signal_filter

    @rf_signal_filter.setter
    def rf_signal_filter(self, value: int):
        self._validate_rf_signal_filter(rf_signal_filter=value)
        self._rf_signal_filter = value

    @staticmethod
    def _validate_lp_rf_timeout(lp_rf_timeout: int) -> None:
        """Validate lp rf timeout parameter.

        Args:
            lp_rf_timeout (int): LP RF timeout.
        Raises:
            RequestParameterInvalidValueError: If lp_rf_timeout is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= lp_rf_timeout <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('LP RF timeout value should be between 0 and 255.')

    @property
    def lp_rf_timeout(self) -> int:
        """:obj:`int`: LP RF timeout.

        Getter and setter.
        """
        return self._lp_rf_timeout

    @lp_rf_timeout.setter
    def lp_rf_timeout(self, value: int):
        self._validate_lp_rf_timeout(lp_rf_timeout=value)
        self._lp_rf_timeout = value

    @staticmethod
    def _validate_uart_baud_rate(uart_baud_rate: Union[dpa_constants.BaudRates, int]) -> None:
        """Validate uart baud rate parameter.

        Args:
            uart_baud_rate (Union[BaudRates, int]): UART Baud rate.
        Raises:
            RequestParameterInvalidValueError: If uart_baud_rate is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= uart_baud_rate <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('UART baud rate value should be between 0 and 255.')

    @property
    def uart_baud_rate(self) -> Union[dpa_constants.BaudRates, int]:
        """:obj:`BaudRates` or :obj:`int`: UART baud rate.

        Getter and setter.
        """
        return self._uart_baud_rate

    @uart_baud_rate.setter
    def uart_baud_rate(self, value: Union[dpa_constants.BaudRates, int]):
        self._validate_uart_baud_rate(uart_baud_rate=value)
        self._uart_baud_rate = value

    @staticmethod
    def _validate_alternative_dsm_channel(alternative_dsm_channels: int) -> None:
        """Validate alternative dsm channel parameters.

        Args:
            alternative_dsm_channels (int): Alternative DSM channel.
        Raises:
            RequestParameterInvalidValueError: If alternative_dsm_channels is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= alternative_dsm_channels <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Alternative DMS channel value should be between 0 and 255.')

    @property
    def alternative_dsm_channel(self) -> int:
        """:obj:`int`: Alternative DSM channel.

        Getter and setter.
        """
        return self._alternative_dsm_channel

    @alternative_dsm_channel.setter
    def alternative_dsm_channel(self, value: int):
        self._validate_alternative_dsm_channel(alternative_dsm_channels=value)
        self._alternative_dsm_channel = value

    @property
    def local_frc(self) -> bool:
        """:obj:`bool`: Local FRC reception enabled.

        Getter and setter.
        """
        return self._local_frc

    @local_frc.setter
    def local_frc(self, value: bool):
        self._local_frc = value

    @staticmethod
    def _validate_rf_channel_a(rf_channel_a: int) -> None:
        """Validate RF Channel A Parameter.

        Args:
            rf_channel_a (int): RF Channel A.
        Raises:
            RequestParameterInvalidValueError: If rf_channel_a is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= rf_channel_a <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('RF channel A value should be between 0 and 255.')

    @property
    def rf_channel_a(self) -> int:
        """:obj:`int`: RF Channel A.

        Getter and setter.
        """
        return self._rf_channel_a

    @rf_channel_a.setter
    def rf_channel_a(self, value: int):
        self._validate_rf_channel_a(rf_channel_a=value)
        self._rf_channel_a = value

    @staticmethod
    def _validate_rf_channel_b(rf_channel_b: int) -> None:
        """Validate RF Channel B Parameter.

        Args:
            rf_channel_b (int): RF Channel B.
        Raises:
            RequestParameterInvalidValueError: If rf_channel_b is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= rf_channel_b <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('RF channel B value should be between 0 and 255.')

    @property
    def rf_channel_b(self) -> int:
        """:obj:`int`: RF Channel B.

        Getter and setter.
        """
        return self._rf_channel_b

    @rf_channel_b.setter
    def rf_channel_b(self, value: int):
        self._validate_rf_channel_b(rf_channel_b=value)
        self._rf_channel_b = value

    @staticmethod
    def _validate_undocumented(undocumented: List[int]) -> None:
        """Validate undocumented data bytes.

        Args:
            undocumented (List[int]): Undocumented data bytes.
        Raises:
            RequestParameterInvalidValueError: If undocumented does not contain 13 values or if values are not
            in range from 0 to 255.
        """
        if len(undocumented) != 13:
            raise RequestParameterInvalidValueError('Undocumented block should be 13B long.')
        if not Common.values_in_byte_range(undocumented):
            raise RequestParameterInvalidValueError('Undocumented block values should be between 0 and 255.')

    @property
    def undocumented(self) -> List[int]:
        """:obj:`list` of :obj:`int`: Undocumented data bytes.

        Getter and setter.
        """
        return self._undocumented

    @undocumented.setter
    def undocumented(self, value: List[int]):
        self._validate_undocumented(undocumented=value)
        self._undocumented = value

    @classmethod
    def from_pdata(cls, data: Union[List[int], bytearray]) -> 'OsTrConfData':
        """Deserialize DPA response pdata into OS TR Configuration data object.

        Returns:
            :obj:`OsTrConfData`: Deserialized OS TR Configuration data object.
        """
        if isinstance(data, bytearray):
            data = list(data)
        embed_pers_data = data[0:4]
        embedded_pers = []
        for i in range(0, len(embed_pers_data * 8)):
            if embed_pers_data[int(i / 8)] & (1 << (i % 8)) and EmbedPeripherals.has_value(i):
                embedded_pers.append(EmbedPeripherals(i))
        embedded_peripherals = embedded_pers
        custom_dpa_handler = bool(data[4] & 1)
        dpa_peer_to_peer = bool(data[4] & 2)
        routing_off = bool(data[4] & 8)
        io_setup = bool(data[4] & 16)
        user_peer_to_peer = bool(data[4] & 32)
        stay_awake_when_not_bonded = bool(data[4] & 64)
        std_and_lp_network = bool(data[4] & 128)
        rf_output_power = data[7]
        rf_signal_filter = data[8]
        lp_rf_timeout = data[9]
        uart_baud_rate = data[10]
        alternative_dsm_channel = data[11]
        local_frc = bool(data[12] & 1)
        rf_channel_a = data[16]
        rf_channel_b = data[17]
        undocumented = data[18:]
        return cls(embedded_peripherals=embedded_peripherals, custom_dpa_handler=custom_dpa_handler,
                   dpa_peer_to_peer=dpa_peer_to_peer, routing_off=routing_off, io_setup=io_setup,
                   user_peer_to_peer=user_peer_to_peer, stay_awake_when_not_bonded=stay_awake_when_not_bonded,
                   std_and_lp_network=std_and_lp_network, rf_output_power=rf_output_power,
                   rf_signal_filter=rf_signal_filter, lp_rf_timeout=lp_rf_timeout,
                   uart_baud_rate=uart_baud_rate, alternative_dsm_channel=alternative_dsm_channel,
                   local_frc=local_frc, rf_channel_a=rf_channel_a, rf_channel_b=rf_channel_b,
                   undocumented=undocumented)

    def to_pdata(self, to_bytes: bool = False) -> Union[List[int], bytearray]:
        """Serialize OS TR Configuration data object to DPA request pdata.

        Args:
            to_bytes (bool): Serialize into bytes.
        Returns:
            :obj:`list` of :obj:`int`: Serialized OS TR Configuration data.\n
            :obj:`bytearray`: Serialize OS TR Configuration data bytes if to_bytes is True.
        """
        embed_pers = Common.peripheral_list_to_bitmap(self.embedded_peripherals)
        conf_bits_0 = int(self.custom_dpa_handler) | int(self.dpa_peer_to_peer) << 1 | int(self.routing_off) << 3 | \
            int(self.io_setup) << 4 | int(self.user_peer_to_peer) << 5 | \
            int(self.stay_awake_when_not_bonded) << 6 | int(self.std_and_lp_network) << 7
        pdata = embed_pers + [conf_bits_0] + [0] * 2 + \
            [self.rf_output_power, self.rf_signal_filter, self.lp_rf_timeout, self.uart_baud_rate,
             self.alternative_dsm_channel, int(self.local_frc)] + [0] * 3 + [self.rf_channel_a, self.rf_channel_b] + \
            self.undocumented
        if to_bytes:
            return bytearray(pdata)
        return pdata



import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.enums.commands import SensorResponseCommands
from iqrfpy.enums.message_types import SensorMessages
from iqrfpy.enums.peripherals import Standards
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.sensor.responses.enumerate import EnumerateResponse
from iqrfpy.utils.quantity_data import get_sensor_class
from iqrfpy.utils.sensor_parser import SensorTypes, SensorData, SensorParser
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': SensorMessages.ENUMERATE,
    'msgid': 'enumerateTest',
    'nadr': 1,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 103,
    'result': {
        'sensors': [
            {
                'id': 'TEMPERATURE',
                'type': 1,
                'name': 'Temperature',
                'shortName': 't',
                'unit': '°C',
                'decimalPlaces': 4,
                'frcs': [
                     144,
                     224
                ]
            },
            {
                'id': 'BINARYDATA7',
                'type': 129,
                'name': 'Binary data7',
                'shortName': 'bin7',
                'unit': '?',
                'decimalPlaces': 0,
                'frcs': [
                     16,
                     144
                ],
                'breakdown': [
                     {
                        'id': 'BINARYDATA7',
                        'type': 129,
                        'name': 'Light indicator',
                        'shortName': 'light',
                        'unit': '%',
                        'decimalPlaces': 1
                     }
                ]
            },
            {
                'id': 'BINARYDATA7',
                'type': 129,
                'name': 'Binary data7',
                'shortName': 'bin7',
                'unit': '?',
                'decimalPlaces': 0,
                'frcs': [
                     16,
                     144
                ],
                'breakdown': [
                     {
                        'id': 'BINARYDATA7',
                        'type': 129,
                        'name': 'Potentiometer',
                        'shortName': 'pot',
                        'unit': '%',
                        'decimalPlaces': 1
                     }
                ]
            }
        ]
    },
    'dpa': b'\x01\x00\x5e\xbe\x02\x04\x00\x67\x01\x81\x81'
}

data_error: dict = {
    'mtype': SensorMessages.ENUMERATE,
    'msgid': 'enumerateTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x5e\xbe\x04\x04\x01\x23'
}


class EnumerateResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, EnumerateResponse.from_dpa(data_ok['dpa']), False],
        ['from_json', data_ok, EnumerateResponse.from_json(generate_json_response(data_ok)), True],
        ['from_dpa_error', data_error, EnumerateResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, EnumerateResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, Standards.SENSOR)
        with self.subTest():
            self.assertEqual(response.pcmd, SensorResponseCommands.ENUMERATE)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, SensorMessages.ENUMERATE)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            EnumerateResponse.from_dpa(b'\x01\x00\x5e\xbe\x02')

    @parameterized.expand([
        [
            'from_dpa',
            [
                SensorTypes.TEMPERATURE,
                SensorTypes.BINARY_DATA7,
                SensorTypes.BINARY_DATA7,
            ],
            EnumerateResponse.from_dpa(data_ok['dpa']),
        ],
        [
            'from_json',
            [
                SensorTypes.TEMPERATURE,
                SensorTypes.BINARY_DATA7,
                SensorTypes.BINARY_DATA7,
            ],
            EnumerateResponse.from_json(generate_json_response(data_ok)),
        ],
    ])
    def test_get_sensor_data(self, _, sensors_types: List[int], response: EnumerateResponse):
        sensor_data = []
        for i in sensors_types:
            sensor_class = get_sensor_class(i)
            data = SensorData(
                sensor_type=sensor_class.type,
                index=len(sensor_data),
                name=sensor_class.name,
                short_name=sensor_class.short_name,
                unit=sensor_class.unit,
                decimal_places=sensor_class.decimal_places,
                frc_commands=sensor_class.frc_commands
            )
            sensor_data.append(data)
        self.assertEqual(response.sensor_data, sensor_data)

    @parameterized.expand([
        ['from_dpa_error', EnumerateResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', EnumerateResponse.from_json(generate_json_response(data_error))],
    ])
    def test_get_sensor_data_error(self, _, response: EnumerateResponse):
        self.assertIsNone(response.sensor_data)



import random
import unittest
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.restore import RestoreRequest


class RestoreRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b''.join(
            [b'\x00\x00\x00\x0c\xff\xff\xed\x60\x14\x51\x82\xde\x7c\x72\xde\xf8\x90\xa3\xee\x82\x75\xed\x90\x54',
             b'\xa7\xb7\x81\x33\xca\x65\x70\x07\x7b\x24\x03\x43\x08\x1e\x82\x62\xce\x0b\x10\x2b\x23\x43\xa3\x58',
             b'\xd6\xa6\x8c\x91\x11\x68\xa8']
        )
        self.json = {
            'mType': 'iqrfEmbedCoordinator_Restore',
            'data': {
                'msgId': 'restoreTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'networkData': [
                            237,
                            96,
                            20,
                            81,
                            130,
                            222,
                            124,
                            114,
                            222,
                            248,
                            144,
                            163,
                            238,
                            130,
                            117,
                            237,
                            144,
                            84,
                            167,
                            183,
                            129,
                            51,
                            202,
                            101,
                            112,
                            7,
                            123,
                            36,
                            3,
                            67,
                            8,
                            30,
                            130,
                            98,
                            206,
                            11,
                            16,
                            43,
                            35,
                            67,
                            163,
                            88,
                            214,
                            166,
                            140,
                            145,
                            17,
                            104,
                            168
                        ]
                    }
                },
                'returnVerbose': True
            }
        }

    def test_to_dpa(self):
        network_data = random.sample(range(0, 255), 49)
        expected = bytes(b'\x00\x00\x00\x0c\xff\xff' + bytes(network_data))
        request = RestoreRequest(network_data=network_data)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    def test_to_json(self):
        network_data = random.sample(range(0, 255), 49)
        request = RestoreRequest(network_data=network_data, msgid='restoreTest')
        self.json['data']['req']['param']['networkData'] = network_data
        self.assertEqual(
            request.to_json(),
            self.json
        )

    def test_set_network_data(self):
        request = RestoreRequest(network_data=self.json['data']['req']['param']['networkData'], msgid='restoreTest')
        print(request.to_dpa())
        print(self.dpa)
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        network_data = random.sample(range(0, 255), 49)
        dpa = bytes(b'\x00\x00\x00\x0c\xff\xff' + bytes(network_data))
        request.network_data = network_data
        self.json['data']['req']['param']['networkData'] = network_data
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [[0] * 60],
        [[-255]],
        [[256]]
    ])
    def test_construct_invalid(self, network_data):
        with self.assertRaises(ValueError):
            RestoreRequest(network_data=network_data)



"""FRC Send Selective request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import FrcRequestCommands
from iqrfpy.enums.message_types import FrcMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['SendSelectiveRequest']


class SendSelectiveRequest(IRequest):
    """FRC Send Selective request class."""

    __slots__ = '_frc_command', '_selected_nodes', '_user_data'

    def __init__(self, nadr: int, frc_command: int, selected_nodes: List[int], user_data: List[int],
                 hwpid: int = dpa_constants.HWPID_MAX, dpa_rsp_time: Optional[float] = None,
                 dev_process_time: Optional[float] = None, msgid: Optional[str] = None):
        """Send Selective request constructor.

        Args:
            nadr (int): Device address.
            frc_command (int): FRC command (data to collect).
            selected_nodes (List[int]): Selected nodes.
            user_data (List[int]): User data (see FRC commands for details).
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(frc_command, selected_nodes, user_data)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.FRC,
            pcmd=FrcRequestCommands.SEND_SELECTIVE,
            m_type=FrcMessages.SEND_SELECTIVE,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._frc_command = frc_command
        self._selected_nodes = selected_nodes
        self._user_data = user_data

    def _validate(self, frc_command: int, selected_nodes: List[int], user_data: List[int]):
        """Validate request parameters.

        Args:
            frc_command (int): FRC command (data to collect).
            selected_nodes (List[int]): Selected nodes.
            user_data (List[int]): User data (see FRC commands for details).
        """
        self._validate_frc_command(frc_command)
        self._validate_selected_nodes(selected_nodes)
        self._validate_user_data(user_data)

    @staticmethod
    def _validate_frc_command(frc_command: int):
        """Validate FRC command parameter.

        Args:
            frc_command (int): FRC command (data to collect).
        Raises:
            RequestParameterInvalidValueError: If frc_command is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= frc_command <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('FRC command value should be between 0 and 255.')

    @property
    def frc_command(self) -> int:
        """:obj:`int`: FRC command.

        Getter and setter.
        """
        return self._frc_command

    @frc_command.setter
    def frc_command(self, value: int):
        self._validate_frc_command(value)
        self._frc_command = value

    @staticmethod
    def _validate_selected_nodes(selected_nodes: List[int]):
        """Validate selected nodes parameter.

        Args:
            selected_nodes (List[int]): Selected nodes.
        Raises:
            RequestParameterInvalidValueError: If selected_nodes contains more than 240 values or values are not
            in range from 0 to 255.
        """
        if len(selected_nodes) > 240:
            raise RequestParameterInvalidValueError('Selected nodes should contain at most 240 values.')
        if min(selected_nodes) < 0 or max(selected_nodes) > 239:
            raise RequestParameterInvalidValueError('Selected nodes values should be between 1 and 239.')

    @property
    def selected_nodes(self) -> List[int]:
        """:obj:`list` of :obj:`int`: Selected nodes.

        Getter and setter.
        """
        return self._selected_nodes

    @selected_nodes.setter
    def selected_nodes(self, value: List[int]):
        self._validate_selected_nodes(value)
        self._selected_nodes = value

    @staticmethod
    def _validate_user_data(user_data: List[int]):
        """Validate user data.

        Args:
            user_data (List[int]): User data (see FRC commands for details).
        Raises:
            RequestParameterInvalidValueError: If user_data contains more than 27 values or values are not
            in range from 0 to 255.
        """
        if len(user_data) > 27:
            raise RequestParameterInvalidValueError('User data should contain at most 27 values.')
        if not Common.values_in_byte_range(user_data):
            raise RequestParameterInvalidValueError('User data values should be between 0 and 255.')

    @property
    def user_data(self) -> List[int]:
        """:obj:`list` of :obj:`int`: User data.

        Getter and setter.
        """
        return self._user_data

    @user_data.setter
    def user_data(self, value: List[int]):
        self._validate_user_data(value)
        self._user_data = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._frc_command] + Common.nodes_to_bitmap(self._selected_nodes) + self._user_data
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {
            'frcCommand': self._frc_command,
            'selectedNodes': self._selected_nodes,
            'userData': self._user_data
        }
        return super().to_json()



"""IO peripheral response messages."""

from .direction import DirectionResponse
from .get import GetResponse
from .set import SetResponse

__all__ = [
    'DirectionResponse',
    'GetResponse',
    'SetResponse'
]



from parameterized import parameterized
import unittest
from typing import List
from iqrfpy.utils.common import *
from iqrfpy.enums.commands import *
from iqrfpy.enums.message_types import *
from iqrfpy.enums.peripherals import *
from iqrfpy.exceptions import JsonMsgidMissingError, JsonDpaValueMissingError, JsonHwpidMissingError, \
    JsonMTypeMissingError, JsonNadrMissingError, JsonRCodeMissingError, JsonResultMissingError, \
    JsonStatusMissingError, UnsupportedMessageTypeError


class CommonTestCase(unittest.TestCase):

    def setUp(self):
        self.request = {
            'mType': 'iqrfEmbedLedr_Set',
            'data': {
                'msgId': 'test',
                'req': {
                    'nAdr': 1,
                    'param': {
                        'onOff': True
                    }
                },
                'returnVerbose': True
            }
        }
        self.response = {
            'mType': 'iqrfEmbedLedr_Set',
            'data': {
                'msgId': 'testEmbedLedr',
                'rsp': {
                    'nAdr': 1,
                    'hwpId': 2,
                    'rCode': 0,
                    'dpaVal': 75,
                    'result': {}
                },
                'raw': {
                    'request': '01.00.06.01.ff.ff',
                    'requestTs': '2018-07-17T16:54:38.112445',
                    'confirmation': '01.00.06.01.ff.ff.ff.3b.01.04.01',
                    'confirmationTs': '2018-07-17T16:54:38.135078',
                    'response': '01.00.06.81.02.00.00.4b',
                    'responseTs': '2018-07-17T16:54:38.277356'
                },
                'insId': 'iqrfgd2-1',
                'statusStr': 'ok',
                'status': 0
            }
        }

    # hwpid_from_dpa tests

    def test_hwpid_from_dpa_ok(self):
        expected = 1026
        self.assertEqual(
            Common.hwpid_from_dpa(0x04, 0x02),
            expected
        )

    def test_hwpid_from_dpa_param_value_too_large(self):
        with self.assertRaises(ValueError):
            Common.hwpid_from_dpa(256, 0)

    def test_hwpid_from_dpa_param_value_negative(self):
        with self.assertRaises(ValueError):
            Common.hwpid_from_dpa(-1, 0)

    # pnum_from_dpa tests

    @parameterized.expand([
        ['RAM', EmbedPeripherals.RAM, EmbedPeripherals.RAM],
        ['OS', 2, EmbedPeripherals.OS],
        ['BinaryOutput', Standards.BINARY_OUTPUT, Standards.BINARY_OUTPUT],
        ['Sensor', 0x5E, Standards.SENSOR]
    ])
    def test_pnum_from_dpa_ok(self, _, value, expected):
        self.assertEqual(
            Common.pnum_from_dpa(value),
            expected
        )

    def test_pnum_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            Common.pnum_from_dpa(256)

    def test_pnum_from_dpa_unknown(self):
        with self.assertRaises(ValueError):
            Common.pnum_from_dpa(128)

    # pcmd_from_dpa

    @parameterized.expand([
        ['Coordinator', EmbedPeripherals.COORDINATOR, 4, CoordinatorRequestCommands.BOND_NODE],
        ['Node', EmbedPeripherals.NODE, 0, NodeRequestCommands.READ],
        ['OS', EmbedPeripherals.OS, OSRequestCommands.INDICATE, OSRequestCommands.INDICATE],
        ['Eeprom', EmbedPeripherals.EEPROM, 0, EEPROMRequestCommands.READ],
        ['Eeeprom', EmbedPeripherals.EEEPROM, 3, EEEPROMRequestCommands.WRITE],
        ['RAM', EmbedPeripherals.RAM, RAMRequestCommands.WRITE, RAMRequestCommands.WRITE],
        ['LEDR', EmbedPeripherals.LEDR, LEDRequestCommands.FLASHING, LEDRequestCommands.FLASHING],
        ['LEDG', EmbedPeripherals.LEDG, 3, LEDRequestCommands.PULSE],
        ['IO', EmbedPeripherals.IO, IORequestCommands.DIRECTION, IORequestCommands.DIRECTION],
        ['Thermometer', EmbedPeripherals.THERMOMETER, ThermometerRequestCommands.READ, ThermometerRequestCommands.READ],
        ['Uart', EmbedPeripherals.UART, 0x03, UartRequestCommands.CLEAR_WRITE_READ],
        ['Frc', EmbedPeripherals.FRC, FrcRequestCommands.EXTRA_RESULT, FrcRequestCommands.EXTRA_RESULT],
        ['Exploration', EmbedPeripherals.EXPLORATION, ExplorationRequestCommands.PERIPHERALS_ENUMERATION_INFORMATION,
         ExplorationRequestCommands.PERIPHERALS_ENUMERATION_INFORMATION],
        ['DALI', Standards.DALI, DALIRequestCommands.SEND_REQUEST_COMMANDS, DALIRequestCommands.SEND_REQUEST_COMMANDS],
        ['BinaryOutput', Standards.BINARY_OUTPUT, 0, BinaryOutputRequestCommands.SET_OUTPUT],
        ['Sensor', Standards.SENSOR, 0x3E, SensorRequestCommands.ENUMERATE],
        ['Light', Standards.LIGHT, LightRequestCommands.INCREMENT_POWER, LightRequestCommands.INCREMENT_POWER]
    ])
    def test_request_pcmd_from_dpa_ok(self, _, pnum, value, expected):
        self.assertEqual(
            Common.request_pcmd_from_dpa(pnum, value),
            expected
        )

    def test_request_pcmd_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            Common.request_pcmd_from_dpa(EmbedPeripherals.OS, 1000)

    def test_request_pcmd_from_dpa_negative(self):
        with self.assertRaises(ValueError):
            Common.request_pcmd_from_dpa(Standards.SENSOR, -1)

    @parameterized.expand([
        ['RAM', EmbedPeripherals.RAM, 5],
        ['BinaryOutput', Standards.BINARY_OUTPUT, 63]
    ])
    def test_request_pcmd_from_dpa_unknown_pcmd(self, _, pnum, value):
        with self.assertRaises(ValueError):
            Common.request_pcmd_from_dpa(pnum, value)

    @parameterized.expand([
        ['Coordinator', EmbedPeripherals.COORDINATOR, 132, CoordinatorResponseCommands.BOND_NODE],
        ['Node', EmbedPeripherals.NODE, 128, NodeResponseCommands.READ],
        ['OS', EmbedPeripherals.OS, OSResponseCommands.INDICATE, OSResponseCommands.INDICATE],
        ['Eeprom', EmbedPeripherals.EEPROM, 128, EEPROMResponseCommands.READ],
        ['Eeeprom', EmbedPeripherals.EEEPROM, 131, EEEPROMResponseCommands.WRITE],
        ['RAM', EmbedPeripherals.RAM, RAMResponseCommands.WRITE, RAMResponseCommands.WRITE],
        ['LEDR', EmbedPeripherals.LEDR, LEDResponseCommands.FLASHING, LEDResponseCommands.FLASHING],
        ['LEDG', EmbedPeripherals.LEDG, 131, LEDResponseCommands.PULSE],
        ['IO', EmbedPeripherals.IO, IOResponseCommands.DIRECTION, IOResponseCommands.DIRECTION],
        ['Thermometer', EmbedPeripherals.THERMOMETER, ThermometerResponseCommands.READ,
         ThermometerResponseCommands.READ],
        ['Uart', EmbedPeripherals.UART, 0x83, UartResponseCommands.CLEAR_WRITE_READ],
        ['Frc', EmbedPeripherals.FRC, FrcResponseCommands.EXTRA_RESULT, FrcResponseCommands.EXTRA_RESULT],
        ['Exploration', EmbedPeripherals.EXPLORATION, ExplorationResponseCommands.PERIPHERALS_ENUMERATION_INFORMATION,
         ExplorationResponseCommands.PERIPHERALS_ENUMERATION_INFORMATION],
        ['DALI', Standards.DALI, DALIResponseCommands.SEND_REQUEST_COMMANDS,
         DALIResponseCommands.SEND_REQUEST_COMMANDS],
        ['BinaryOutput', Standards.BINARY_OUTPUT, 128, BinaryOutputResponseCommands.SET_OUTPUT],
        ['Sensor', Standards.SENSOR, 0xBE, SensorResponseCommands.ENUMERATE],
        ['Light', Standards.LIGHT, LightResponseCommands.INCREMENT_POWER, LightResponseCommands.INCREMENT_POWER]
    ])
    def test_response_pcmd_from_dpa_ok(self, _, pnum, value, expected):
        self.assertEqual(
            Common.response_pcmd_from_dpa(pnum, value),
            expected
        )

    def test_response_pcmd_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            Common.response_pcmd_from_dpa(EmbedPeripherals.OS, 1000)

    def test_response_pcmd_from_dpa_negative(self):
        with self.assertRaises(ValueError):
            Common.response_pcmd_from_dpa(Standards.SENSOR, -1)

    @parameterized.expand([
        ['RAM', EmbedPeripherals.RAM, 133],
        ['BinaryOutput', Standards.BINARY_OUTPUT, 240]
    ])
    def test_response_pcmd_from_dpa_unknown_pcmd(self, _, pnum, value):
        with self.assertRaises(ValueError):
            Common.response_pcmd_from_dpa(pnum, value)

    # msgid_from_json tests

    def test_msgid_from_json_ok(self):
        expected = 'test'
        self.assertEqual(
            Common.msgid_from_json(self.request),
            expected
        )

    @parameterized.expand([
        [{}],
        [{'data': {'req': {}}}]
    ])
    def test_msgid_from_json_missing_key(self, value):
        with self.assertRaises(JsonMsgidMissingError):
            Common.msgid_from_json(value)

    # mtype_str_from_json tests

    def test_mtype_str_from_json_ok(self):
        self.assertEqual(
            Common.mtype_str_from_json(self.response),
            'iqrfEmbedLedr_Set'
        )

    @parameterized.expand([
        [{}],
        [{'data': {'req': {}}}]
    ])
    def test_mtype_str_from_json_missing_key(self, value):
        with self.assertRaises(JsonMTypeMissingError):
            Common.mtype_str_from_json(value)

    # nadr_from_json tests

    def test_nadr_from_json_ok(self):
        expected = 1
        self.assertEqual(
            Common.nadr_from_json(self.response),
            expected,
        )

    @parameterized.expand([
        [{}],
        [{'data': {'req': {}}}]
    ])
    def test_nadr_from_json_missing_key(self, value):
        with self.assertRaises(JsonNadrMissingError):
            Common.nadr_from_json(value)

    # hwpid_from_json tests

    def test_hwpid_from_json_ok(self):
        expected = 2
        self.assertEqual(
            Common.hwpid_from_json(self.response),
            expected
        )

    @parameterized.expand([
        [{}],
        [{'data': {'req': {}}}]
    ])
    def test_hwpid_from_json_missing_key(self, value):
        with self.assertRaises(JsonHwpidMissingError):
            Common.hwpid_from_json(value)

    # rcode_from_json tests

    def test_rcode_from_json_ok(self):
        expected = 0
        self.assertEqual(
            Common.rcode_from_json(self.response),
            expected
        )

    @parameterized.expand([
        [{}],
        [{'data': {'req': {}}}]
    ])
    def test_rcode_from_json_missing_key(self, value):
        with self.assertRaises(JsonRCodeMissingError):
            Common.rcode_from_json(value)

    # dpa_value_from_json tests

    def test_dpa_value_from_json_ok(self):
        expected = 75
        self.assertEqual(
            Common.dpa_value_from_json(self.response),
            expected
        )

    @parameterized.expand([
        [{}],
        [{'data': {'req': {}}}]
    ])
    def test_dpa_value_from_json_missing_key(self, value):
        with self.assertRaises(JsonDpaValueMissingError):
            Common.dpa_value_from_json(value)

    # result_from_json tests

    def test_result_from_json_ok(self):
        expected = {}
        self.assertEqual(
            Common.result_from_json(self.response),
            expected
        )

    @parameterized.expand([
        [{}],
        [{'data': {'req': {}}}]
    ])
    def test_result_from_json_missing_key(self, value):
        with self.assertRaises(JsonResultMissingError):
            Common.result_from_json(value)

    def test_status_from_json_ok(self):
        expected = 0
        self.assertEqual(
            Common.status_from_json(self.response),
            expected
        )

    @parameterized.expand([
        [{}],
        [{'data': {'req': {}}}]
    ])
    def test_status_from_json_missing_key(self, value):
        with self.assertRaises(JsonStatusMissingError):
            Common.status_from_json(value)

    # string_to_mtype tests

    @parameterized.expand([
        ['Raw', 'iqrfRaw', GenericMessages.RAW],
        ['AddrInfo', 'iqrfEmbedCoordinator_AddrInfo', CoordinatorMessages.ADDR_INFO],
        ['SensorRead', 'iqrfSensor_ReadSensorsWithTypes', SensorMessages.READ_SENSORS_WITH_TYPES],
        ['LightEnumerate', 'iqrfLight_Enumerate', LightMessages.ENUMERATE]
    ])
    def test_string_to_mtype_ok(self, _, string, expected):
        self.assertEqual(Common.string_to_mtype(string), expected)

    @parameterized.expand(['test', 'unknown', 'iqrfEmbedCoordinator_Nonexistent'])
    def test_string_to_mtype_unknown(self, string):
        with self.assertRaises(UnsupportedMessageTypeError):
            Common.string_to_mtype(string)

    # bitmap_to_nodes tests

    @parameterized.expand([
        ['empty', [], False, []],
        ['single', [0x0e], True, [1, 2, 3]],
        ['full', [0xFF, 0xFF, 0xFF, 0xFF], True, [i for i in range(1, 32)]],
        ['coordinator_no_shift', [0xFF], False, [i for i in range(0, 8)]],
        ['coordinator_shift', [0xFF], True, [i for i in range(1, 8)]]
    ])
    def test_bitmap_to_nodes(self, _, bitmap: List[int], shift: bool, expected: List[int]):
        self.assertEqual(
            Common.bitmap_to_nodes(bitmap, shift),
            expected
        )

    # nodes_to_bitmap tests

    @parameterized.expand([
        ['empty', [], [0] * 30],
        ['single', [1, 2, 3], [0x0e] + [0] * 29],
        ['full', [i for i in range(0, 32)], [0xFF, 0xFF, 0xFF, 0xFF] + [0] * 26]
    ])
    def test_nodes_to_bitmap(self, _, nodes, expected):
        self.assertEqual(
            Common.nodes_to_bitmap(nodes),
            expected
        )

    # is_hex_string tests

    @parameterized.expand([
        ['empty', '', False],
        ['hexadecimal', '96bcac11a', True],
        ['non-hexadecimal', 'abcdefx', False]
    ])
    def test_is_hex_string(self, _, value, expected):
        self.assertEqual(
            Common.is_hex_string(value),
            expected
        )

    # hex_string_to_list tests

    @parameterized.expand([
        ['short', '0efa', [14, 250]],
        ['ibk', '9a691f1a2101216503e5c588ffe7f6c2', [154, 105, 31, 26, 33, 1, 33, 101,
                                                     3, 229, 197, 136, 255, 231, 246, 194]]
    ])
    def test_hex_string_to_list_ok(self, _, string, expected):
        self.assertEqual(
            Common.hex_string_to_list(string),
            expected
        )

    @parameterized.expand([
        ['empty', ''],
        ['odd_length_string', 'aab'],
        ['non-hexadecimal', 'aaax']
    ])
    def test_hex_string_to_list_invalid(self, _, value):
        with self.assertRaises(ValueError):
            Common.hex_string_to_list(value)

    # values_in_byte_range tests

    @parameterized.expand([
        ['empty', [], True],
        ['bytes', [255, 10, 12, 7], True],
        ['too_high', [1000, 5], False],
        ['negative', [-1, 10, 12, 7], False]
    ])
    def test_values_in_byte_range(self, _, value, expected):
        self.assertEqual(
            Common.values_in_byte_range(value),
            expected
        )



"""Exceptions module.

This module contains exceptions classes for library-wide use.
"""

from typing import Optional

__all__ = [
    'RequestNadrInvalidError',
    'RequestPnumInvalidError',
    'RequestPcmdInvalidError',
    'RequestHwpidInvalidError',
    'RequestParameterInvalidValueError',
    'DpaConfirmationPacketError',
    'DpaConfirmationPacketLengthError',
    'DpaResponsePacketLengthError',
    'JsonMsgidMissingError',
    'JsonMTypeMissingError',
    'JsonNadrMissingError',
    'JsonHwpidMissingError',
    'JsonRCodeMissingError',
    'JsonDpaValueMissingError',
    'JsonResultMissingError',
    'JsonStatusMissingError',
    'InvalidPeripheralValueError',
    'InvalidPeripheralCommandValueError',
    'UnsupportedPeripheralError',
    'UnsupportedPeripheralCommandError',
    'UnsupportedMessageTypeError',
    'TransportNotConnectedError',
    'DpaRequestTimeoutError',
    'JsonRequestTimeoutError',
    'JsonResponseNotSupportedError',
    'JsonGenericDataMissingError',
    'MessageNotReceivedError',
    'UnknownSensorTypeError',
    'ResponseFactoryRegisterError',
]


class RequestNadrInvalidError(ValueError):
    """Invalid NADR value.

    This error is raised whenever a request object receives
    address that is out of allowed range ([C]: 0, [N]: 1-239).
    """


class RequestPnumInvalidError(ValueError):
    """Invalid PNUM value.

    This error is raised whenever a request object receives
    PNUM that is out of allowed range (0-255 / 0x00-0xFF).
    """


class RequestPcmdInvalidError(ValueError):
    """Invalid PCMD value.

    This error is raised whenever a request object receives
    PCMD that is out of allowed range (0-255 / 0x00-0xFF).
    """


class RequestHwpidInvalidError(ValueError):
    """Invalid HWPID value.

    This error is raised whenever a request object receives
    HWPID that is out of allowed range (0-65535 / 0x0000-0xFFFF).
    """


class RequestParameterInvalidValueError(ValueError):
    """Invalid request parameter value.

    This error is raised whenever a request object receives
    a parameter with value outside of it's defined domain.
    """


class DpaConfirmationPacketError(ValueError):
    """Invalid DPA confirmation packet.

    This error is raised whenever a packet does not meet requirements
    of a DPA confirmation packet.
    """


class DpaConfirmationPacketLengthError(ValueError):
    """Invalid DPA confirmation packet length.

    This error is raised whenever a DPA confirmation packet length does
    not match expected packet length.
    """


class DpaResponsePacketLengthError(ValueError):
    """Invalid DPA response packet length.

    This error is raised whenever a DPA response packet length
    does not match the expected packet length of a specific DPA response.
    """


class JsonMsgidMissingError(KeyError):
    """msgId property missing.

    This error is raised whenever a json response object
    does not contain the msgId property.
    """


class JsonMTypeMissingError(KeyError):
    """mType property missing.

    This error is raised whenever a json response object
    does not contain the mType property.
    """


class JsonNadrMissingError(KeyError):
    """nAdr property missing.

    This error is raised whenever a json response object
    does not contain the nAdr property.
    """


class JsonHwpidMissingError(KeyError):
    """hwpId property missing.

    This error is raised whenever a json response object
    does not contain the hwpId property.
    """


class JsonRCodeMissingError(KeyError):
    """rCode property missing.

    This error is raised whenever a json response object
    does not contain the rCode property.
    """


class JsonDpaValueMissingError(KeyError):
    """dpaVal property missing.

    This error is raised whenever a json response object
    does not contain the dpaVal property.
    """


class JsonResultMissingError(KeyError):
    """result property missing.

    This error is raised whenever a json response object
    does not contain the result property.
    """


class JsonStatusMissingError(KeyError):
    """status property missing.

    This error is raised whenever a json response object
    does not contain the status property.
    """


class JsonGenericDataMissingError(KeyError):
    """rData property missing.

    This error is raised when a generic json response object
    does not contain the rData property.
    """


class InvalidPeripheralValueError(ValueError):
    """Invalid peripheral value.

    This error is raised whenever a peripheral parameter value is out
    of defined range.
    """


class InvalidPeripheralCommandValueError(ValueError):
    """Invalid peripheral command value.

    This error is raised whenever a peripheral command parameter value is out
    of defined range.
    """


class UnsupportedPeripheralError(ValueError):
    """Peripheral unknown or unsupported.

    This error is raised whenever a peripheral parameter value is valid,
    but not recognized as a known and supported peripheral.
    """


class UnsupportedPeripheralCommandError(ValueError):
    """Peripheral command unknown or unsupported.

    This error is raised whenever a peripheral parameter command value is valid,
    but not recognized as a known and supported peripheral command.
    """


class UnsupportedMessageTypeError(ValueError):
    """Message type unknown or unsupported.

    This error is raised whenever a message type parameter is used
    to determine a value, but the message type parameter value is not
    recognized as a known and supported message type.
    """

    def __init__(self, message: str, msgid: Optional[str] = None):
        """UnsupportedMessageTypeError constructor.

        Args:
            message (str): Error message.
            msgid (str, optional): Message ID. Defaults to None.
        """
        super().__init__(message)
        self.msgid = msgid


class TransportNotConnectedError(ConnectionError):
    """Transport not connected.

    This error is raised whenever a transport object is tasked
    with sending a message while not being connected or having lost
    connection.
    """


class DpaRequestTimeoutError(TimeoutError):
    """DPA request not handled or response not received.

    This error is raised whenever transport does not receive a DPA response
    within specified time. If JSON API is used, a JSON response is received,
    but the contents indicate DPA timeout.
    """


class JsonRequestTimeoutError(TimeoutError):
    """JSON API response not received.

    This error is raised whenever transport does not receive a JSON API
    response within specified time.
    """


class JsonResponseNotSupportedError(ValueError):
    """JSON API response not supported.

    This error is raised whenever transport receives a response message that is not supported,
    but the message ID matches a request sent prior to the fact.
    """


class MessageNotReceivedError(TimeoutError):
    """Message not received.

    This error is raised during parsing of a JSON API response
    indicating that DPA response was not received.
    """

    def __init__(self, message: str, msgid: Optional[str] = None):
        """MessageNotReceivedError constructor.

        Args:
            message (str): Error message.
            msgid (str, optional): Message ID. Defaults to None.
        """
        super().__init__(message)
        self.msgid = msgid


class UnknownSensorTypeError(ValueError):
    """Unknown or unsupported sensor type.

    This error is raised when parsing Sensor data collected using regular or FRC requests
    and the sensor type value is not recognized by the library.
    """


class ResponseFactoryRegisterError(KeyError):
    """Response factory cannot be registered.

    This error is raised when a response factory is attempted to be registered under existing
    combination of pnum and pcmd for DPA messages or existing message type for JSON API responses.
    """



"""Confirmation module.

This module contains DPA Confirmation message class.
"""

from typing import List, Optional
from iqrfpy.enums.commands import Command
from iqrfpy.enums.peripherals import Peripheral
from iqrfpy.utils.common import Common
from iqrfpy.utils.dpa import ResponsePacketMembers
from iqrfpy.iresponse import IResponseGetterMixin
from iqrfpy.utils.validators import DpaValidator

__all__ = ['Confirmation']


class Confirmation(IResponseGetterMixin):
    """Confirmation message class.

    DPA confirmation is used to confirm reception of DPA request by node device at coordinator.
    The message carries DPA value, request hops (number of hops used to deliver DPA request to node device),
    response hops (number of hops used to deliver DPA response from node device to coordinator),
    and timeslot (see DPA documentation for timeslot length calculation).
    """

    __slots__ = '_request_hops', '_response_hops', '_timeslot'

    def __init__(self, nadr: int, pnum: Peripheral, pcmd: Command, hwpid: int, dpa_value: int, rcode: int,
                 pdata: Optional[List[int]] = None, result: Optional[dict] = None):
        """Confirmation constructor.

        Args:
            nadr (int): Device address.
            pnum (Peripheral): Peripheral.
            pcmd (Command): Peripheral command.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            rcode (int, optional): Response code. Defaults to 128.
            dpa_value (int, optional): DPA value. Defaults to 0.
            pdata (List[int], optional): DPA response data. Defaults to None.
            result (dict, optional): JSON response data. Defaults to None.
        """
        super().__init__(
            nadr=nadr,
            pcmd=pcmd,
            pnum=pnum,
            hwpid=hwpid,
            rcode=rcode,
            dpa_value=dpa_value,
            pdata=pdata,
            result=result
        )
        self._request_hops: int = result['requestHops']
        self._response_hops: int = result['responseHops']
        self._timeslot: int = result['timeslot']

    @property
    def request_hops(self) -> int:
        """:obj:`int`: Request hops.

        Getter only.
        """
        return self._request_hops

    @property
    def response_hops(self) -> int:
        """:obj:`int`: Response hops.

        Getter only.
        """
        return self._response_hops

    @property
    def timeslot(self) -> int:
        """:obj:`int`: Timeslot.

        Getter only.
        """
        return self._timeslot

    @classmethod
    def from_dpa(cls, dpa: bytes) -> 'Confirmation':
        """Confirmation DPA factory method.

        Parses DPA confirmation message and constructs :obj:`Confirmation` object.

        Args:
            dpa (bytes): DPA confirmation bytes.
        Returns:
            :obj:`Confirmation`: Confirmation message object.
        """
        DpaValidator.confirmation_length(dpa=dpa)
        DpaValidator.confirmation_code(dpa=dpa)
        nadr = dpa[ResponsePacketMembers.NADR]
        pnum = Common.pnum_from_dpa(dpa[ResponsePacketMembers.PNUM])
        pcmd = Common.request_pcmd_from_dpa(pnum, dpa[ResponsePacketMembers.PCMD])
        hwpid = Common.hwpid_from_dpa(dpa[ResponsePacketMembers.HWPID_HI], dpa[ResponsePacketMembers.HWPID_LO])
        rcode = dpa[ResponsePacketMembers.RCODE]
        dpa_value = dpa[ResponsePacketMembers.DPA_VALUE]
        pdata = list(dpa[8:])
        result = {'requestHops': dpa[8], 'responseHops': dpa[10], 'timeslot': dpa[9]}
        return cls(nadr=nadr, pnum=pnum, pcmd=pcmd, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value,
                   pdata=pdata, result=result)

    @classmethod
    def from_json(cls, json: dict) -> 'Confirmation':
        """Confirmation JSON factory method.

        This method is not implemented as JSON API does not support standalone Confirmation messages.

        Raises:
            NotImplementedError: This method is not implemented.
        """
        raise NotImplementedError('from_json() method not implemented.')



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import ExplorationResponseCommands
from iqrfpy.enums.message_types import ExplorationMessages
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.exploration.responses.peripheral_information import PeripheralInformationResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': ExplorationMessages.PERIPHERAL_INFORMATION,
    'msgid': 'peripheralInformationTest',
    'pnum': 2,
    'nadr': 3,
    'hwpid': 5122,
    'rcode': 0,
    'dpa_value': 79,
    'result': {
        'perTe': 3,
        'perT': 3,
        'par1': 3,
        'par2': 169,
    },
    'dpa': b'\x03\x00\x02\xbf\x02\x14\x00\x4f\x03\x03\x03\xa9'
}

data_error: dict = {
    'mtype': ExplorationMessages.PERIPHERAL_INFORMATION,
    'msgid': 'peripheralInformationTest',
    'pnum': 2,
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 7,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x02\xbf\x04\x04\x07\x23'
}


class PeripheralInformationResponseTestCase(unittest.TestCase):
    @parameterized.expand([
        ['from_dpa', data_ok, PeripheralInformationResponse.from_dpa(data_ok['dpa']), False],
        ['from_json', data_ok, PeripheralInformationResponse.from_json(generate_json_response(data_ok)), True],
        ['from_dpa_error', data_error, PeripheralInformationResponse.from_dpa(data_error['dpa']), False],
        [
            'from_json_error',
            data_error,
            PeripheralInformationResponse.from_json(
                generate_json_response(data_error)
            ),
            True,
        ]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, response_data['pnum'])
        with self.subTest():
            self.assertEqual(response.pcmd, ExplorationResponseCommands.PERIPHERALS_ENUMERATION_INFORMATION)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, ExplorationMessages.PERIPHERAL_INFORMATION)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            PeripheralInformationResponse.from_dpa(b'\x03\x00\x02\xbf\x02\x14\x00\x4f\x03')

    @parameterized.expand([
        ['from_dpa', data_ok['result']['perTe'], PeripheralInformationResponse.from_dpa(data_ok['dpa'])],
        [
            'from_json',
            data_ok['result']['perTe'],
            PeripheralInformationResponse.from_json(
                generate_json_response(data_ok)
            )
        ]
    ])
    def test_get_perte(self, _, perte: int, response: PeripheralInformationResponse):
        self.assertEqual(response.peripheral_data.perte, perte)

    @parameterized.expand([
        ['from_dpa', data_ok['result']['perT'], PeripheralInformationResponse.from_dpa(data_ok['dpa'])],
        [
            'from_json',
            data_ok['result']['perT'],
            PeripheralInformationResponse.from_json(
                generate_json_response(data_ok)
            )
        ]
    ])
    def test_get_pert(self, _, pert: int, response: PeripheralInformationResponse):
        self.assertEqual(response.peripheral_data.pert, pert)

    @parameterized.expand([
        ['from_dpa', data_ok['result']['par1'], PeripheralInformationResponse.from_dpa(data_ok['dpa'])],
        [
            'from_json',
            data_ok['result']['par1'],
            PeripheralInformationResponse.from_json(
                generate_json_response(data_ok)
            )
        ]
    ])
    def test_get_par1(self, _, par1: int, response: PeripheralInformationResponse):
        self.assertEqual(response.peripheral_data.par1, par1)

    @parameterized.expand([
        ['from_dpa', data_ok['result']['par2'], PeripheralInformationResponse.from_dpa(data_ok['dpa'])],
        [
            'from_json',
            data_ok['result']['par2'],
            PeripheralInformationResponse.from_json(
                generate_json_response(data_ok)
            )
        ]
    ])
    def test_get_par2(self, _, par2: int, response: PeripheralInformationResponse):
        self.assertEqual(response.peripheral_data.par2, par2)

    @parameterized.expand([
        ['from_dpa_error', PeripheralInformationResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', PeripheralInformationResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_peripheral_data_error(self, _, response: PeripheralInformationResponse):
        self.assertIsNone(response.peripheral_data)



"""IO peripheral request messages."""

from .direction import DirectionRequest
from .get import GetRequest
from .set import SetRequest
from .io_triplet import IoTriplet

__all__ = [
    'IoTriplet',
    'DirectionRequest',
    'GetRequest',
    'SetRequest',
]



import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.enums.commands import UartResponseCommands
from iqrfpy.enums.message_types import UartMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.uart.responses.write_read import WriteReadResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': UartMessages.WRITE_READ,
    'msgid': 'writeReadTest',
    'nadr': 3,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 90,
    'readData': [1, 2, 3, 4, 5],
    'result': {
        'readData': [1, 2, 3, 4, 5],
    },
    'dpa': b'\x03\x00\x0c\x82\x00\x00\x00\x5a\x01\x02\x03\x04\x05'
}

data_ok_1: dict = {
    'mtype': UartMessages.WRITE_READ,
    'msgid': 'writeReadTest',
    'nadr': 3,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'readData': [10, 5, 1, 17, 2],
    'result': {
        'readData': [10, 5, 1, 17, 2],
    },
    'dpa': b'\x03\x00\x0c\x82\x02\x04\x00\x23\x0a\x05\x01\x11\x02'
}

data_error: dict = {
    'mtype': UartMessages.WRITE_READ,
    'msgid': 'writeReadTest',
    'nadr': 1,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x01\x00\x0c\x82\x04\x04\x01\x23'
}


class WriteReadResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, WriteReadResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, WriteReadResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, WriteReadResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, WriteReadResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, WriteReadResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, WriteReadResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.UART)
        with self.subTest():
            self.assertEqual(response.pcmd, UartResponseCommands.WRITE_READ)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, UartMessages.WRITE_READ)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            WriteReadResponse.from_dpa(b'\x03\x00\x0c\x82\x00')

    @parameterized.expand([
        ['from_dpa', data_ok['readData'], WriteReadResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['readData'], WriteReadResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['readData'], WriteReadResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['readData'], WriteReadResponse.from_json(generate_json_response(data_ok_1))],
    ])
    def test_get_read_data(self, _, read_data: List[int], response: WriteReadResponse):
        self.assertEqual(response.read_data, read_data)

    @parameterized.expand([
        ['from_dpa_error', WriteReadResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', WriteReadResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_read_data_error(self, _, response: WriteReadResponse):
        self.assertIsNone(response.read_data)



"""Coordinator Discovery response message."""

from typing import List, Optional
from iqrfpy.iresponse import IResponseGetterMixin
from iqrfpy.enums.commands import CoordinatorResponseCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.utils.dpa import ResponseCodes, ResponsePacketMembers
from iqrfpy.utils.validators import DpaValidator, JsonValidator

__all__ = ['DiscoveryResponse']


class DiscoveryResponse(IResponseGetterMixin):
    """Coordinator Discovery response class."""

    __slots__ = ('_disc_nr',)

    def __init__(self, hwpid: int = dpa_constants.HWPID_MAX, rcode: int = 0, dpa_value: int = 0,
                 msgid: Optional[str] = None, pdata: Optional[List[int]] = None, result: Optional[dict] = None):
        """Discovery response constructor.

        Args:
            hwpid (int, optional): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            rcode (int, optional): Response code. Defaults to 128.
            dpa_value (int, optional): DPA value. Defaults to 0.
            pdata (List[int], optional): DPA response data. Defaults to None.
            msgid (str, optional): Message ID. Defaults to None.
            result (dict, optional): JSON response data. Defaults to None.
        """
        super().__init__(
            nadr=dpa_constants.COORDINATOR_NADR,
            pnum=EmbedPeripherals.COORDINATOR,
            pcmd=CoordinatorResponseCommands.DISCOVERY,
            m_type=CoordinatorMessages.DISCOVERY,
            hwpid=hwpid,
            rcode=rcode,
            dpa_value=dpa_value,
            msgid=msgid,
            pdata=pdata,
            result=result
        )
        self._disc_nr = result['discNr'] if rcode == ResponseCodes.OK else None

    @property
    def disc_nr(self) -> Optional[int]:
        """:obj:`int` or :obj:`None`: Number of discovered devices.

        Getter only.
        """
        return self._disc_nr

    @classmethod
    def from_dpa(cls, dpa: bytes) -> 'DiscoveryResponse':
        """DPA response factory method.

        Parses DPA data and constructs DiscoveryResponse object.

        Args:
            dpa (bytes): DPA response bytes.
        Returns:
            DiscoveryResponse: Response message object.
        """
        DpaValidator.base_response_length(dpa=dpa)
        hwpid = Common.hwpid_from_dpa(dpa[ResponsePacketMembers.HWPID_HI], dpa[ResponsePacketMembers.HWPID_LO])
        rcode = dpa[ResponsePacketMembers.RCODE]
        dpa_value = dpa[ResponsePacketMembers.DPA_VALUE]
        pdata = None
        result = None
        if rcode == ResponseCodes.OK:
            DpaValidator.response_length(dpa=dpa, expected_len=9)
            pdata = Common.pdata_from_dpa(dpa=dpa)
            result = {'discNr': dpa[8]}
        return cls(hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata, result=result)

    @classmethod
    def from_json(cls, json: dict) -> 'DiscoveryResponse':
        """JSON response factory method.

        Parses JSON API response and constructs DiscoveryResponse object.

        Args:
            json (dict): JSON API response.
        Returns:
            DiscoveryResponse: Response message object.
        """
        JsonValidator.response_received(json=json)
        msgid = Common.msgid_from_json(json=json)
        hwpid = Common.hwpid_from_json(json=json)
        dpa_value = Common.dpa_value_from_json(json=json)
        rcode = Common.rcode_from_json(json=json)
        pdata = Common.pdata_from_json(json=json)
        result = Common.result_from_json(json=json) if rcode == ResponseCodes.OK else None
        return cls(msgid=msgid, hwpid=hwpid, dpa_value=dpa_value, rcode=rcode, pdata=pdata, result=result)



"""OS Read Data object."""

from dataclasses import dataclass
from iqrfpy.peripherals.exploration.responses.peripheral_enumeration_data import PeripheralEnumerationData


@dataclass
class OsReadData:
    """OS Read Data class."""

    __slots__ = 'mid', 'os_version', 'tr_type', 'os_build', 'rssi', 'supply_voltage', 'flags', 'slot_limits', 'ibk', \
        'per_enum'

    def __init__(self, data: dict):
        """Read Data constructor.

        Args:
            data (dict): OS Read data.
        """
        self.mid: int = data['mid']
        """Module ID."""
        self.os_version: int = data['osVersion']
        """OS version."""
        self.tr_type: int = data['trMcuType']
        """MCU type and TR series."""
        self.os_build: int = data['osBuild']
        """OS build."""
        self.rssi: int = data['rssi']
        """RSSI."""
        self.supply_voltage: int = data['supplyVoltage']
        """Supply voltage."""
        self.flags: int = data['flags']
        """OS flags."""
        self.slot_limits: int = data['slotLimits']
        """Slot limits."""
        self.ibk: int = data['ibk']
        """Individual bonding key."""
        self.per_enum: PeripheralEnumerationData = PeripheralEnumerationData(data)
        """Peripheral enumeration data."""



"""FRC Params parameters."""

from dataclasses import dataclass
from iqrfpy.utils.dpa import FrcResponseTimes


@dataclass
class FrcParams:
    """FRC Params class."""

    __slots__ = 'offline_frc', 'frc_response_time'

    def __init__(self, offline_frc: bool = False, frc_response_time: FrcResponseTimes = FrcResponseTimes.MS40):
        """FRC Params constructor.

        Args:
            offline_frc (bool): Offline FRC.
            frc_response_time (FrcResponseTimes): FRC response time.
        """
        self.offline_frc: bool = offline_frc
        """Offline FRC."""
        self.frc_response_time: FrcResponseTimes = frc_response_time
        """FRC response time."""

    @classmethod
    def from_int(cls, value: int) -> 'FrcParams':
        """Deserializes FRC Params integer value into FRC Params object.

        Args:
            value (int): FRC params integer value.
        """
        return cls(
            offline_frc=bool(value & 0x08),
            frc_response_time=FrcResponseTimes(value & 0x70)
        )

    def to_data(self) -> int:
        """Serialize FRC Params object into an integer value.

        Returns:
            :obj:`int`: Serialized FRC params value.
        """
        return self.frc_response_time | int(self.offline_frc) << 3



import random
import unittest
from dataclasses import dataclass
from typing import Union
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.smart_connect import SmartConnectRequest


@dataclass
class DataRegular:
    req_addr = 10
    bonding_test_retries = 3
    ibk = [32, 28, 94, 118, 47, 117, 215, 226, 68, 195, 250, 9, 179, 222, 14, 161]
    mid = 2165321045
    virtual_device_address = 255
    expected_dpa = b''.join([
        b'\x00\x00\x00\x12\xff\xff\x0a\x03\x20\x1c\x5e\x76\x2f\x75\xd7\xe2\x44\xc3\xfa\x09\xb3\xde\x0e\xa1\x55'
        b'\x2d\x10\x81\x00\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    ])


class DataTempAddr:
    req_addr = 254
    bonding_test_retries = 0
    ibk = [111, 249, 157, 155, 80, 108, 206, 40, 171, 71, 174, 231, 200, 33, 83, 146]
    mid = 0
    virtual_device_address = 0
    expected_dpa = b''.join([
        b'\x00\x00\x00\x12\xff\xff\xfe\x00\x6f\xf9\x9d\x9b\x50\x6c\xce\x28\xab\x47\xae\xe7\xc8\x21\x53\x92\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    ])


class SmartConnectRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b''.join([
            b'\x00\x00\x00\x12\xff\xff\x0b\x01\x9a\x69\x1f\x1a\x21\x01\x21\x65\x03\xe5\xc5\x88\xff\xe7\xf6\xc2\x55',
            b'\x2d\x10\x81\x00\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ])
        self.json = {
            'mType': 'iqrfEmbedCoordinator_SmartConnect',
            'data': {
                'msgId': 'smartConnectTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'reqAddr': 11,
                        'bondingTestRetries': 1,
                        'ibk': [
                            154,
                            105,
                            31,
                            26,
                            33,
                            1,
                            33,
                            101,
                            3,
                            229,
                            197,
                            136,
                            255,
                            231,
                            246,
                            194
                        ],
                        'mid': 2165321045,
                        'virtualDeviceAddress': 255,
                        'userData': [
                            0,
                            0,
                            0,
                            0
                        ]
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['regular', DataRegular],
        ['temporary_addr', DataTempAddr]
    ])
    def test_to_dpa(self, _, data: Union[DataRegular, DataTempAddr]):
        request = SmartConnectRequest(
            req_addr=data.req_addr,
            bonding_test_retries=data.bonding_test_retries,
            ibk=data.ibk,
            mid=data.mid,
            virtual_device_address=data.virtual_device_address
        )
        print(request.to_dpa())
        print(data.expected_dpa)
        self.assertEqual(
            request.to_dpa(),
            data.expected_dpa
        )

    @parameterized.expand([
        ['regular', DataRegular],
        ['temporary_addr', DataTempAddr]
    ])
    def test_to_json(self, _, data: Union[DataRegular, DataTempAddr]):
        request = SmartConnectRequest(
            req_addr=data.req_addr,
            bonding_test_retries=data.bonding_test_retries,
            ibk=data.ibk,
            mid=data.mid,
            virtual_device_address=data.virtual_device_address,
            msgid='smartConnectTest'
        )
        self.json['data']['req']['param']['reqAddr'] = data.req_addr
        self.json['data']['req']['param']['bondingTestRetries'] = data.bonding_test_retries
        self.json['data']['req']['param']['ibk'] = data.ibk
        self.json['data']['req']['param']['mid'] = data.mid
        self.json['data']['req']['param']['virtualDeviceAddress'] = data.virtual_device_address
        print(request.to_json())
        print(self.json)
        self.assertEqual(
            request.to_json(),
            self.json
        )

    def test_set_req_addr(self):
        request = SmartConnectRequest(
            req_addr=self.json['data']['req']['param']['reqAddr'],
            bonding_test_retries=self.json['data']['req']['param']['bondingTestRetries'],
            ibk=self.json['data']['req']['param']['ibk'],
            mid=self.json['data']['req']['param']['mid'],
            virtual_device_address=self.json['data']['req']['param']['virtualDeviceAddress'],
            msgid='smartConnectTest'
        )
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        request.req_addr = DataRegular.req_addr
        self.json['data']['req']['param']['reqAddr'] = DataRegular.req_addr
        dpa = list(self.dpa)
        dpa[6] = DataRegular.req_addr
        dpa = bytes(dpa)
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    def test_set_bonding_test_retries(self):
        request = SmartConnectRequest(
            req_addr=self.json['data']['req']['param']['reqAddr'],
            bonding_test_retries=self.json['data']['req']['param']['bondingTestRetries'],
            ibk=self.json['data']['req']['param']['ibk'],
            mid=self.json['data']['req']['param']['mid'],
            virtual_device_address=self.json['data']['req']['param']['virtualDeviceAddress'],
            msgid='smartConnectTest'
        )
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        request.bonding_test_retries = DataRegular.bonding_test_retries
        self.json['data']['req']['param']['bondingTestRetries'] = DataRegular.bonding_test_retries
        dpa = list(self.dpa)
        dpa[7] = DataRegular.bonding_test_retries
        dpa = bytes(dpa)
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    def test_set_ibk(self):
        request = SmartConnectRequest(
            req_addr=self.json['data']['req']['param']['reqAddr'],
            bonding_test_retries=self.json['data']['req']['param']['bondingTestRetries'],
            ibk=self.json['data']['req']['param']['ibk'],
            mid=self.json['data']['req']['param']['mid'],
            virtual_device_address=self.json['data']['req']['param']['virtualDeviceAddress'],
            msgid='smartConnectTest'
        )
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        request.ibk = DataRegular.ibk
        self.json['data']['req']['param']['ibk'] = DataRegular.ibk
        dpa = list(self.dpa)
        dpa = dpa[:8] + DataRegular.ibk + dpa[24:]
        dpa = bytes(dpa)
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    def test_set_mid(self):
        request = SmartConnectRequest(
            req_addr=self.json['data']['req']['param']['reqAddr'],
            bonding_test_retries=self.json['data']['req']['param']['bondingTestRetries'],
            ibk=self.json['data']['req']['param']['ibk'],
            mid=self.json['data']['req']['param']['mid'],
            virtual_device_address=self.json['data']['req']['param']['virtualDeviceAddress'],
            msgid='smartConnectTest'
        )
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        request.mid = DataRegular.mid
        self.json['data']['req']['param']['mid'] = DataRegular.mid
        dpa_mid = [
            DataRegular.mid & 0xFF,
            (DataRegular.mid >> 8) & 0xFF,
            (DataRegular.mid >> 16) & 0xFF,
            (DataRegular.mid >> 24) & 0xFF
        ]
        dpa = list(self.dpa)
        dpa = dpa[:24] + dpa_mid + dpa[28:]
        dpa = bytes(dpa)
        print(request.to_dpa())
        print(dpa)
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    def test_set_virtual_device_address(self):
        request = SmartConnectRequest(
            req_addr=self.json['data']['req']['param']['reqAddr'],
            bonding_test_retries=self.json['data']['req']['param']['bondingTestRetries'],
            ibk=self.json['data']['req']['param']['ibk'],
            mid=self.json['data']['req']['param']['mid'],
            virtual_device_address=self.json['data']['req']['param']['virtualDeviceAddress'],
            msgid='smartConnectTest'
        )
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        request.virtual_device_address = DataRegular.virtual_device_address
        self.json['data']['req']['param']['virtualDeviceAddress'] = DataRegular.virtual_device_address
        dpa = list(self.dpa)
        dpa[29] = DataRegular.virtual_device_address
        dpa = bytes(dpa)
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        ['req_addr', -1, 0, [random.randint(0, 255)] * 16, 0, 0],
        ['req_addr', 275, 0, [random.randint(0, 255)] * 16, 0, 0],
        ['bonding_test_retries', 1, -1, [random.randint(0, 255)] * 16, 0, 0],
        ['bonding_test_retries', 1, 270, [random.randint(0, 255)] * 16, 0, 0],
        ['ibk', 1, 1, [10], 0, 0],
        ['ibk', 1, 1, [random.randint(0, 255)] * 17, 0, 0],
        ['ibk', 1, 1, [0, 0, 0, 0, 0, 0, 0, 0, 256, 0, 0, 0, 0, 0, 0, 0], 0, 0],
        ['mid', 1, 1, [random.randint(0, 255)] * 16, -1, 0],
        ['mid', 1, 1, [random.randint(0, 255)] * 16, 4294967298, 0],
        ['virtual_device_address', 1, 1, [random.randint(0, 255)] * 16, 0, -1],
        ['virtual_device_address', 1, 1, [random.randint(0, 255)] * 16, 0, 256],
    ])
    def test_construct_invalid(self, _, req_addr, bonding_test_retries, ibk, mid, virtual_device_address):
        with self.assertRaises(ValueError):
            SmartConnectRequest(
                req_addr=req_addr,
                bonding_test_retries=bonding_test_retries,
                ibk=ibk,
                mid=mid,
                virtual_device_address=virtual_device_address
            )



"""Coordinator peripheral request messages.

As the address of the coordinator device is always 0, the Coordinator
request messages do not provide the device address parameter (nadr),
and automatically assign address 0.
"""

from .addr_info import AddrInfoRequest
from .authorize_bond import AuthorizeBondRequest, AuthorizeBondParams
from .backup import BackupRequest
from .bond_node import BondNodeRequest
from .bonded_devices import BondedDevicesRequest
from .clear_all_bonds import ClearAllBondsRequest
from .discovered_devices import DiscoveredDevicesRequest
from .discovery import DiscoveryRequest
from .remove_bond import RemoveBondRequest
from .restore import RestoreRequest
from .set_dpa_params import SetDpaParamsRequest, DpaParam
from .set_hops import SetHopsRequest
from .set_mid import SetMidRequest
from .smart_connect import SmartConnectRequest

__all__ = [
    'AddrInfoRequest',
    'AuthorizeBondRequest',
    'AuthorizeBondParams',
    'BackupRequest',
    'BondNodeRequest',
    'BondedDevicesRequest',
    'ClearAllBondsRequest',
    'DiscoveredDevicesRequest',
    'DiscoveryRequest',
    'RemoveBondRequest',
    'RestoreRequest',
    'SetDpaParamsRequest',
    'DpaParam',
    'SetHopsRequest',
    'SetMidRequest',
    'SmartConnectRequest'
]



import unittest
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.addr_info import AddrInfoRequest
from iqrfpy.peripherals.node.requests.read import ReadRequest


class IRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.request = AddrInfoRequest(msgid='addrInfoTest')
        self.json = {
            'mType': 'iqrfEmbedCoordinator_AddrInfo',
            'data': {
                'msgId': 'test',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {},
                },
                'returnVerbose': True
            },
        }

    @parameterized.expand([
        ['immutable', False, b'\x00\x00\x00\x00\xff\xff'],
        ['mutable', True, bytearray(b'\x00\x00\x00\x00\xff\xff')]
    ])
    def test_to_dpa(self, _, mutable, expected):
        self.assertEqual(
            self.request.to_dpa(mutable),
            expected
        )

    @parameterized.expand([
        ['nadr_low', -1, 65535],
        ['nadr_high', 1000, 65535],
        ['hwpid_low', 1, -10],
        ['hwpid_high', 1, 100000]
    ])
    def test_validate_base_invalid(self, _, nadr, hwpid):
        with self.assertRaises(ValueError):
            ReadRequest(nadr=nadr, hwpid=hwpid)
